#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------


cdef extern from "azure_c_shared_utility/strings_types.h":

    ctypedef struct STRING_HANDLE:
        pass


cdef extern from "azure_c_shared_utility/strings.h":

    STRING_HANDLE STRING_new()
    STRING_HANDLE STRING_clone(STRING_HANDLE handle)
    STRING_HANDLE STRING_construct(const char* psz)
    STRING_HANDLE STRING_construct_n(const char* psz, size_t n)
    STRING_HANDLE STRING_new_with_memory(const char* memory)
    STRING_HANDLE STRING_new_quoted(const char* source)
    STRING_HANDLE STRING_new_JSON(const char* source)
    STRING_HANDLE STRING_from_byte_array(const unsigned char* source, size_t size)
    void STRING_delete(STRING_HANDLE handle)
    int STRING_concat(STRING_HANDLE handle, const char* s2)
    int STRING_concat_with_STRING(STRING_HANDLE s1, STRING_HANDLE s2)
    int STRING_quote(STRING_HANDLE handle)
    int STRING_copy(STRING_HANDLE s1, const char* s2)
    int STRING_copy_n(STRING_HANDLE s1, const char* s2, size_t n)
    const char* STRING_c_str(STRING_HANDLE handle)
    int STRING_empty(STRING_HANDLE handle)
    size_t STRING_length(STRING_HANDLE handle)
    int STRING_compare(STRING_HANDLE s1, STRING_HANDLE s2)
    int STRING_replace(STRING_HANDLE handle, char target, char replace)



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

from libc cimport stdint

cimport c_amqpvalue


cdef extern from "azure_uamqp_c/amqp_definitions.h":

    ctypedef stdint.uint_fast32_t tickcounter_ms_t

    # role
    ctypedef bint role
    c_amqpvalue.AMQP_VALUE amqpvalue_create_role(role value)
    cdef bint role_sender
    cdef bint role_receiver
    #cdef c_amqpvalue.amqpvalue_get_boolean amqpvalue_get_role

    # sender-settle-mode
    ctypedef stdint.uint8_t sender_settle_mode
    c_amqpvalue.AMQP_VALUE amqpvalue_create_sender_settle_mode(sender_settle_mode value)
    cdef stdint.uint8_t sender_settle_mode_unsettled
    cdef stdint.uint8_t sender_settle_mode_settled
    cdef stdint.uint8_t sender_settle_mode_mixed
    #cdef amqpvalue_get_sender_settle_mode

    # receiver-settle-mode
    ctypedef stdint.uint8_t receiver_settle_mode
    c_amqpvalue.AMQP_VALUE amqpvalue_create_receiver_settle_mode(receiver_settle_mode value)
    cdef int receiver_settle_mode_first
    cdef int receiver_settle_mode_second
    #cdef amqpvalue_get_receiver_settle_mode

    # handle
    ctypedef stdint.uint32_t handle
    c_amqpvalue.AMQP_VALUE amqpvalue_create_handle(handle value)
    #cdef amqpvalue_get_handle
    # seconds
    ctypedef stdint.uint32_t seconds
    c_amqpvalue.AMQP_VALUE amqpvalue_create_seconds(seconds value)
    #cdef amqpvalue_get_seconds

    # milliseconds
    ctypedef stdint.uint32_t milliseconds
    c_amqpvalue.AMQP_VALUE amqpvalue_create_milliseconds(milliseconds value)
    #cdef amqpvalue_get_milliseconds

    # delivery-tag 
    ctypedef c_amqpvalue.amqp_binary delivery_tag
    c_amqpvalue.AMQP_VALUE amqpvalue_create_delivery_tag(delivery_tag value)
    #cdef amqpvalue_get_delivery_tag

    # sequence-no
    ctypedef stdint.uint32_t sequence_no
    c_amqpvalue.AMQP_VALUE amqpvalue_create_sequence_no(sequence_no value)
    #cdef amqpvalue_get_sequence_no

    # delivery-number
    ctypedef sequence_no delivery_number
    c_amqpvalue.AMQP_VALUE amqpvalue_create_delivery_number(delivery_number value)
    #cdef amqpvalue_get_delivery_number

    # transfer-number
    ctypedef sequence_no transfer_number
    c_amqpvalue.AMQP_VALUE amqpvalue_create_transfer_number(transfer_number value)
    #cdef amqpvalue_get_transfer_number

    # message-format
    ctypedef stdint.uint32_t message_format
    c_amqpvalue.AMQP_VALUE amqpvalue_create_message_format(message_format value)
    #cdef amqpvalue_get_message_format

    # ietf-language-tag
    ctypedef char* ietf_language_tag
    c_amqpvalue.AMQP_VALUE amqpvalue_create_ietf_language_tag(ietf_language_tag value)
    #cdef amqpvalue_get_ietf_language_tag

    # fields
    ctypedef c_amqpvalue.AMQP_VALUE fields
    c_amqpvalue.AMQP_VALUE amqpvalue_create_fields(c_amqpvalue.AMQP_VALUE value)
    #cdef fields_clone
    #cdef fields_destroy
    #cdef amqpvalue_get_fields

    # annotations
    ctypedef c_amqpvalue.AMQP_VALUE annotations
    c_amqpvalue.AMQP_VALUE amqpvalue_create_annotations(c_amqpvalue.AMQP_VALUE value)
    # cdef amqpvalue_get_annotations c_amqpvalue.amqpvalue_get_map  TODO


    # message-id-ulong
    ctypedef stdint.uint64_t message_id_ulong
    c_amqpvalue.AMQP_VALUE amqpvalue_create_message_id_ulong(message_id_ulong value)
    #cdef amqpvalue_get_message_id_ulong


    # message-id-uuid
    ctypedef c_amqpvalue.uuid message_id_uuid
    c_amqpvalue.AMQP_VALUE amqpvalue_create_message_id_uuid(message_id_uuid value)
    #cdef amqpvalue_get_message_id_uuid


    # message-id-binary
    ctypedef c_amqpvalue.amqp_binary message_id_binary
    c_amqpvalue.AMQP_VALUE amqpvalue_create_message_id_binary(message_id_binary value)
    #cdef amqpvalue_get_message_id_binary


    # message-id-string
    ctypedef char* message_id_string
    c_amqpvalue.AMQP_VALUE amqpvalue_create_message_id_string(message_id_string value)
    #cdef amqpvalue_get_message_id_string


    # address-string
    ctypedef char* address_string
    c_amqpvalue.AMQP_VALUE amqpvalue_create_address_string(address_string value)
    #cdef amqpvalue_get_address_string

    # error
    ctypedef struct ERROR_HANDLE:
        pass

    ERROR_HANDLE error_create(const char* condition_value)
    ERROR_HANDLE error_clone(ERROR_HANDLE value)
    void error_destroy(ERROR_HANDLE error)
    int error_get_condition(ERROR_HANDLE error, const char** condition_value)
    int error_set_condition(ERROR_HANDLE error, const char* condition_value)
    int error_get_description(ERROR_HANDLE error, const char** description_value)
    int error_set_description(ERROR_HANDLE error, const char* description_value)
    int error_get_info(ERROR_HANDLE error, fields* info_value)
    int error_set_info(ERROR_HANDLE error, fields info_value)

    # header
    ctypedef struct HEADER_HANDLE:
        pass

    HEADER_HANDLE header_create()
    HEADER_HANDLE header_clone(HEADER_HANDLE value)
    void header_destroy(HEADER_HANDLE header)
    bint is_header_type_by_descriptor(c_amqpvalue.AMQP_VALUE descriptor)
    int amqpvalue_get_header(c_amqpvalue.AMQP_VALUE value, HEADER_HANDLE* HEADER_handle)
    c_amqpvalue.AMQP_VALUE amqpvalue_create_header(HEADER_HANDLE header)
    int header_get_durable(HEADER_HANDLE header, bint* durable_value)
    int header_set_durable(HEADER_HANDLE header, bint durable_value)
    int header_get_priority(HEADER_HANDLE header, stdint.uint8_t* priority_value)
    int header_set_priority(HEADER_HANDLE header, stdint.uint8_t priority_value)
    int header_get_ttl(HEADER_HANDLE header, milliseconds* ttl_value)
    int header_set_ttl(HEADER_HANDLE header, milliseconds ttl_value)
    int header_get_first_acquirer(HEADER_HANDLE header, bint* first_acquirer_value)
    int header_set_first_acquirer(HEADER_HANDLE header, bint first_acquirer_value)
    int header_get_delivery_count(HEADER_HANDLE header, stdint.uint32_t* delivery_count_value)
    int header_set_delivery_count(HEADER_HANDLE header, stdint.uint32_t delivery_count_value)

    # delivery-annotations
    ctypedef annotations delivery_annotations
    c_amqpvalue.AMQP_VALUE amqpvalue_create_delivery_annotations(delivery_annotations value)
    bint is_delivery_annotations_type_by_descriptor(c_amqpvalue.AMQP_VALUE descriptor)
    #cdef amqpvalue_get_delivery_annotations amqpvalue_get_annotations TODO

    # message-annotations
    ctypedef annotations message_annotations
    c_amqpvalue.AMQP_VALUE amqpvalue_create_message_annotations(message_annotations value)
    bint is_message_annotations_type_by_descriptor(c_amqpvalue.AMQP_VALUE descriptor)
    #cdef amqpvalue_get_message_annotations amqpvalue_get_annotations TODO

    # application-properties
    ctypedef c_amqpvalue.AMQP_VALUE application_properties
    c_amqpvalue.AMQP_VALUE amqpvalue_create_application_properties(c_amqpvalue.AMQP_VALUE value)
    #cdef application_properties_clone
    #cdef application_properties_destroy
    bint is_application_properties_type_by_descriptor(c_amqpvalue.AMQP_VALUE descriptor)
    #cdef amqpvalue_get_application_properties

    # data
    ctypedef c_amqpvalue.amqp_binary data
    c_amqpvalue.AMQP_VALUE amqpvalue_create_data(data value)
    bint is_data_type_by_descriptor(c_amqpvalue.AMQP_VALUE descriptor)
    #cdef amqpvalue_get_data

    # amqp-sequence
    ctypedef c_amqpvalue.AMQP_VALUE amqp_sequence
    c_amqpvalue.AMQP_VALUE amqpvalue_create_amqp_sequence(c_amqpvalue.AMQP_VALUE value)
    #cdef amqp_sequence_clone
    #cdef amqp_sequence_destroy
    bint is_amqp_sequence_type_by_descriptor(c_amqpvalue.AMQP_VALUE descriptor)
    #cdef amqpvalue_get_amqp_sequence

    # amqp-value
    ctypedef c_amqpvalue.AMQP_VALUE amqp_value
    c_amqpvalue.AMQP_VALUE amqpvalue_create_amqp_value(c_amqpvalue.AMQP_VALUE value)
    #cdef amqp_value_clone
    #cdef amqp_value_destroy
    bint is_amqp_value_type_by_descriptor(c_amqpvalue.AMQP_VALUE descriptor)
    #cdef amqpvalue_get_amqp_value

    # footer
    ctypedef annotations footer
    c_amqpvalue.AMQP_VALUE amqpvalue_create_footer(footer value)
    bint is_footer_type_by_descriptor(c_amqpvalue.AMQP_VALUE descriptor)
    #cdef amqpvalue_get_footer

    # properties
    ctypedef struct PROPERTIES_HANDLE:
        pass
    PROPERTIES_HANDLE properties_create()
    PROPERTIES_HANDLE properties_clone(PROPERTIES_HANDLE value)
    void properties_destroy(PROPERTIES_HANDLE properties)
    bint is_properties_type_by_descriptor(c_amqpvalue.AMQP_VALUE descriptor)
    int amqpvalue_get_properties(c_amqpvalue.AMQP_VALUE value, PROPERTIES_HANDLE* PROPERTIES_handle)
    c_amqpvalue.AMQP_VALUE amqpvalue_create_properties(PROPERTIES_HANDLE properties)
    int properties_get_message_id(PROPERTIES_HANDLE properties, c_amqpvalue.AMQP_VALUE* message_id_value)
    int properties_set_message_id(PROPERTIES_HANDLE properties, c_amqpvalue.AMQP_VALUE message_id_value)
    int properties_get_user_id(PROPERTIES_HANDLE properties, c_amqpvalue.amqp_binary* user_id_value)
    int properties_set_user_id(PROPERTIES_HANDLE properties, c_amqpvalue.amqp_binary user_id_value)
    int properties_get_to(PROPERTIES_HANDLE properties, c_amqpvalue.AMQP_VALUE* to_value)
    int properties_set_to(PROPERTIES_HANDLE properties, c_amqpvalue.AMQP_VALUE to_value)
    int properties_get_subject(PROPERTIES_HANDLE properties, char** subject_value)
    int properties_set_subject(PROPERTIES_HANDLE properties, char* subject_value)
    int properties_get_reply_to(PROPERTIES_HANDLE properties, c_amqpvalue.AMQP_VALUE* reply_to_value)
    int properties_set_reply_to(PROPERTIES_HANDLE properties, c_amqpvalue.AMQP_VALUE reply_to_value)
    int properties_get_correlation_id(PROPERTIES_HANDLE properties, c_amqpvalue.AMQP_VALUE* correlation_id_value)
    int properties_set_correlation_id(PROPERTIES_HANDLE properties, c_amqpvalue.AMQP_VALUE correlation_id_value)
    int properties_get_content_type(PROPERTIES_HANDLE properties, char** content_type_value)
    int properties_set_content_type(PROPERTIES_HANDLE properties, char* content_type_value)
    int properties_get_content_encoding(PROPERTIES_HANDLE properties, char** content_encoding_value)
    int properties_set_content_encoding(PROPERTIES_HANDLE properties, char* content_encoding_value)
    int properties_get_absolute_expiry_time(PROPERTIES_HANDLE properties, c_amqpvalue.timestamp* absolute_expiry_time_value)
    int properties_set_absolute_expiry_time(PROPERTIES_HANDLE properties, c_amqpvalue.timestamp absolute_expiry_time_value)
    int properties_get_creation_time(PROPERTIES_HANDLE properties, c_amqpvalue.timestamp* creation_time_value)
    int properties_set_creation_time(PROPERTIES_HANDLE properties, c_amqpvalue.timestamp creation_time_value)
    int properties_get_group_id(PROPERTIES_HANDLE properties, char** group_id_value)
    int properties_set_group_id(PROPERTIES_HANDLE properties, char* group_id_value)
    int properties_get_group_sequence(PROPERTIES_HANDLE properties, sequence_no* group_sequence_value)
    int properties_set_group_sequence(PROPERTIES_HANDLE properties, sequence_no group_sequence_value)
    int properties_get_reply_to_group_id(PROPERTIES_HANDLE properties, char** reply_to_group_id_value)
    int properties_set_reply_to_group_id(PROPERTIES_HANDLE properties, char* reply_to_group_id_value)


    # source/target types
    ctypedef stdint.uint32_t terminus_durability
    cdef stdint.uint32_t terminus_durability_none
    cdef stdint.uint32_t terminus_durability_configuration
    cdef stdint.uint32_t terminus_durability_unsettled_state

    ctypedef const char* terminus_expiry_policy
    cdef const char* terminus_expiry_policy_link_detach
    cdef const char* terminus_expiry_policy_session_end
    cdef const char* terminus_expiry_policy_connection_close
    cdef const char* terminus_expiry_policy_never

    ctypedef fields node_properties
    ctypedef c_amqpvalue.AMQP_VALUE filter_set  # Map

    # source
    ctypedef struct SOURCE_HANDLE:
        pass

    SOURCE_HANDLE source_create()
    SOURCE_HANDLE source_clone(SOURCE_HANDLE value)
    void source_destroy(SOURCE_HANDLE source)
    bint is_source_type_by_descriptor(c_amqpvalue.AMQP_VALUE descriptor)
    int amqpvalue_get_source(c_amqpvalue.AMQP_VALUE value, SOURCE_HANDLE* SOURCE_handle)
    c_amqpvalue.AMQP_VALUE amqpvalue_create_source(SOURCE_HANDLE source)

    int source_get_address(SOURCE_HANDLE source, c_amqpvalue.AMQP_VALUE* address_value)
    int source_set_address(SOURCE_HANDLE source, c_amqpvalue.AMQP_VALUE address_value)
    int source_get_durable(SOURCE_HANDLE source, terminus_durability* durable_value)
    int source_set_durable(SOURCE_HANDLE source, terminus_durability durable_value)
    int source_get_expiry_policy(SOURCE_HANDLE source, terminus_expiry_policy* expiry_policy_value)
    int source_set_expiry_policy(SOURCE_HANDLE source, terminus_expiry_policy expiry_policy_value)
    int source_get_timeout(SOURCE_HANDLE source, seconds* timeout_value)
    int source_set_timeout(SOURCE_HANDLE source, seconds timeout_value)
    int source_get_dynamic(SOURCE_HANDLE source, bint* dynamic_value)
    int source_set_dynamic(SOURCE_HANDLE source, bint dynamic_value)
    int source_get_dynamic_node_properties(SOURCE_HANDLE source, node_properties* dynamic_node_properties_value)
    int source_set_dynamic_node_properties(SOURCE_HANDLE source, node_properties dynamic_node_properties_value)
    int source_get_distribution_mode(SOURCE_HANDLE source, const char** distribution_mode_value)
    int source_set_distribution_mode(SOURCE_HANDLE source, const char* distribution_mode_value)
    int source_get_filter(SOURCE_HANDLE source, filter_set* filter_value)
    int source_set_filter(SOURCE_HANDLE source, filter_set filter_value)
    #int source_get_default_outcome(SOURCE_HANDLE source, c_amqpvalue.AMQP_VALUE* default_outcome_value)
    #int source_set_default_outcome(SOURCE_HANDLE source, c_amqpvalue.AMQP_VALUE default_outcome_value)
    #int source_get_outcomes(SOURCE_HANDLE source, c_amqpvalue.AMQP_VALUE* outcomes_value)
    #int source_set_outcomes(SOURCE_HANDLE source, c_amqpvalue.AMQP_VALUE outcomes_value)
    #int source_get_capabilities(SOURCE_HANDLE source, c_amqpvalue.AMQP_VALUE* capabilities_value)
    #int source_set_capabilities(SOURCE_HANDLE source, c_amqpvalue.AMQP_VALUE capabilities_value)

    # target
    ctypedef struct TARGET_HANDLE:
        pass

    TARGET_HANDLE target_create()
    TARGET_HANDLE target_clone(TARGET_HANDLE value)
    void target_destroy(TARGET_HANDLE target)
    bint is_target_type_by_descriptor(c_amqpvalue.AMQP_VALUE descriptor)
    int amqpvalue_get_target(c_amqpvalue.AMQP_VALUE value, TARGET_HANDLE* TARGET_handle)
    c_amqpvalue.AMQP_VALUE amqpvalue_create_target(TARGET_HANDLE target)

    int target_get_address(TARGET_HANDLE target, c_amqpvalue.AMQP_VALUE* address_value)
    int target_set_address(TARGET_HANDLE target, c_amqpvalue.AMQP_VALUE address_value)
    int target_get_durable(TARGET_HANDLE target, terminus_durability* durable_value)
    int target_set_durable(TARGET_HANDLE target, terminus_durability durable_value)
    int target_get_expiry_policy(TARGET_HANDLE target, terminus_expiry_policy* expiry_policy_value)
    int target_set_expiry_policy(TARGET_HANDLE target, terminus_expiry_policy expiry_policy_value)
    int target_get_timeout(TARGET_HANDLE target, seconds* timeout_value)
    int target_set_timeout(TARGET_HANDLE target, seconds timeout_value)
    int target_get_dynamic(TARGET_HANDLE target, bint* dynamic_value)
    int target_set_dynamic(TARGET_HANDLE target, bint dynamic_value)
    int target_get_dynamic_node_properties(TARGET_HANDLE target, node_properties* dynamic_node_properties_value)
    int target_set_dynamic_node_properties(TARGET_HANDLE target, node_properties dynamic_node_properties_value)
    #int target_get_capabilities(TARGET_HANDLE target, c_amqpvalue.AMQP_VALUE* capabilities_value)
    #int target_set_capabilities(TARGET_HANDLE target, c_amqpvalue.AMQP_VALUE capabilities_value)



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

# Python imports
from enum import Enum
import logging
import io

# C imports
from libc.stdlib cimport malloc, free
cimport c_xlogging


_logger = logging.getLogger('uamqp.c_uamqp')


class LogCategory(Enum):
    Error = c_xlogging.LOG_CATEGORY_TAG.AZ_LOG_ERROR
    Info = c_xlogging.LOG_CATEGORY_TAG.AZ_LOG_INFO
    Debug = c_xlogging.LOG_CATEGORY_TAG.AZ_LOG_TRACE


cdef char* vprintf_alloc(const char* format, c_xlogging.va_list va):
    cdef char* result
    cdef int neededSize
    neededSize = c_xlogging.vsnprintf(NULL, 0, format, va)
    if neededSize < 0:
        result = NULL
    else:
        result = <char*>malloc(neededSize + 1)
        if <void*>result != NULL:
            if c_xlogging.vsnprintf(result, neededSize + 1, format, va) != neededSize:
                free(result)
                result = NULL
    return result;


cdef void custom_logging_function(c_xlogging.LOG_CATEGORY_TAG log_category, const char* file, const char* func, const int line, unsigned int options, const char* format, ...):
    log_level = LogCategory(log_category)
    cdef c_xlogging.va_list args
    cdef char* text
    c_xlogging.va_start(args, format)
    try:
        text = vprintf_alloc(format, args)
        if <void*>text != NULL:
            _python_log(log_level, text, bool(options), file=file, func=func, line=line)
    except KeyboardInterrupt:
        pass
    finally:
        c_xlogging.va_end(args)
        if <void*>text != NULL:
            free(text)


cpdef set_python_logger():
    c_xlogging.xlogging_set_log_function(<c_xlogging.LOGGER_LOG>custom_logging_function)


def _python_log(category, text, end, text_bldr=[], file=None, func=None, line=None):
    text_bldr.append(text)
    if not end:
        return

    log_line = b''
    while text_bldr:
        log_line += text_bldr.pop(0)
    if category == LogCategory.Debug or category == LogCategory.Info:
        _logger.info("%r", log_line)
    else:
        _logger.info("%r (%r:%r:%r)", log_line, file, func, line)



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

import logging

from uamqp import constants, errors, sender
from uamqp.utils import get_running_loop

_logger = logging.getLogger(__name__)


class MessageSenderAsync(sender.MessageSender):
    """An asynchronous Message Sender that opens its own exclsuive Link on an
    existing Session.

    :ivar link_credit: The sender Link credit that determines how many
     messages the Link will attempt to handle per connection iteration.
    :vartype link_credit: int
    :ivar properties: Metadata to be sent in the Link ATTACH frame.
    :vartype properties: dict
    :ivar send_settle_mode: The mode by which to settle message send
     operations. If set to `Unsettled`, the client will wait for a confirmation
     from the service that the message was successfully send. If set to 'Settled',
     the client will not wait for confirmation and assume success.
    :vartype send_settle_mode: ~uamqp.constants.SenderSettleMode
    :ivar receive_settle_mode: The mode by which to settle message receive
     operations. If set to `PeekLock`, the receiver will lock a message once received until
     the client accepts or rejects the message. If set to `ReceiveAndDelete`, the service
     will assume successful receipt of the message and clear it from the queue. The
     default is `PeekLock`.
    :vartype receive_settle_mode: ~uamqp.constants.ReceiverSettleMode
    :ivar max_message_size: The maximum allowed message size negotiated for the Link.
    :vartype max_message_size: int

    :param session: The underlying Session with which to send.
    :type session: ~uamqp.async_ops.session_async.SessionAsync
    :param source: The name of source (i.e. the client).
    :type source: str or bytes
    :param target: The AMQP endpoint to send to.
    :type target: ~uamqp.address.Target
    :param name: A unique name for the sender. If not specified a GUID will be used.
    :type name: str or bytes
    :param send_settle_mode: The mode by which to settle message send
     operations. If set to `Unsettled`, the client will wait for a confirmation
     from the service that the message was successfully send. If set to 'Settled',
     the client will not wait for confirmation and assume success.
    :type send_settle_mode: ~uamqp.constants.SenderSettleMode
    :param receive_settle_mode: The mode by which to settle message receive
     operations. If set to `PeekLock`, the receiver will lock a message once received until
     the client accepts or rejects the message. If set to `ReceiveAndDelete`, the service
     will assume successful receipt of the message and clear it from the queue. The
     default is `PeekLock`.
    :type receive_settle_mode: ~uamqp.constants.ReceiverSettleMode
    :param max_message_size: The maximum allowed message size negotiated for the Link.
    :type max_message_size: int
    :param link_credit: The sender Link credit that determines how many
     messages the Link will attempt to handle per connection iteration.
    :type link_credit: int
    :param properties: Metadata to be sent in the Link ATTACH frame.
    :type properties: dict
    :param error_policy: A policy for parsing errors on link, connection and message
     disposition to determine whether the error should be retryable.
    :type error_policy: ~uamqp.errors.ErrorPolicy
    :param debug: Whether to turn on network trace logs. If `True`, trace logs
     will be logged at INFO level. Default is `False`.
    :type debug: bool
    :param encoding: The encoding to use for parameters supplied as strings.
     Default is 'UTF-8'
    :type encoding: str
    :param loop: A user specified event loop.
    :type loop: ~asycnio.AbstractEventLoop
    """

    def __init__(self, session, source, target,
                 name=None,
                 send_settle_mode=constants.SenderSettleMode.Unsettled,
                 receive_settle_mode=constants.ReceiverSettleMode.PeekLock,
                 max_message_size=constants.MAX_MESSAGE_LENGTH_BYTES,
                 link_credit=None,
                 properties=None,
                 error_policy=None,
                 debug=False,
                 encoding='UTF-8',
                 loop=None):
        self.loop = loop or get_running_loop()
        super(MessageSenderAsync, self).__init__(
            session, source, target,
            name=name,
            send_settle_mode=send_settle_mode,
            receive_settle_mode=receive_settle_mode,
            max_message_size=max_message_size,
            link_credit=link_credit,
            properties=properties,
            error_policy=error_policy,
            debug=debug,
            encoding=encoding)

    async def __aenter__(self):
        """Open the MessageSender in an async context manager."""
        await self.open_async()
        return self

    async def __aexit__(self, *args):
        """Close the MessageSender when exiting an async context manager."""
        await self.destroy_async()

    async def destroy_async(self):
        """Asynchronously close both the Sender and the Link. Clean up any C objects."""
        self.destroy()

    async def open_async(self):
        """Asynchronously open the MessageSender in order to start
        processing messages.

        :raises: ~uamqp.errors.AMQPConnectionError if the Sender raises
         an error on opening. This can happen if the target URI is invalid
         or the credentials are rejected.
        """
        try:
            self._sender.open()
        except ValueError:
            raise errors.AMQPConnectionError(
                "Failed to open Message Sender. "
                "Please confirm credentials and target URI.")

    async def send_async(self, message, callback, timeout=0):
        """Add a single message to the internal pending queue to be processed
        by the Connection without waiting for it to be sent.

        :param message: The message to send.
        :type message: ~uamqp.message.Message
        :param callback: The callback to be run once a disposition is received
         in receipt of the message. The callback must take three arguments, the message,
         the send result and the optional delivery condition (exception).
        :type callback:
         callable[~uamqp.message.Message, ~uamqp.constants.MessageSendResult, ~uamqp.errors.MessageException]
        :param timeout: An expiry time for the message added to the queue. If the
         message is not sent within this timeout it will be discarded with an error
         state. If set to 0, the message will not expire. The default is 0.
        """
        # pylint: disable=protected-access
        try:
            raise self._error
        except TypeError:
            pass
        except Exception as e:
            _logger.warning("%r", e)
            raise
        c_message = message.get_message()
        message._on_message_sent = callback
        try:
            await self._session._connection.lock_async(timeout=None)
            return self._sender.send(c_message, timeout, message)
        finally:
            self._session._connection.release_async()

    async def close_async(self):
        """Close the sender asynchronously, leaving the link intact."""
        self._sender.close()



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

# Python imports
import logging

# C imports
cimport c_xio
cimport c_wsio
cimport c_sasl_mechanism


_logger = logging.getLogger(__name__)


cpdef xio_from_wsioconfig(WSIOConfig io_config):
    cdef const c_xio.IO_INTERFACE_DESCRIPTION* ws_io_interface
    ws_io_interface = c_wsio.wsio_get_interface_description()
    xio = XIO()
    xio.create(ws_io_interface, &io_config._c_value)
    return xio


cpdef xio_from_tlsioconfig(IOInterfaceDescription io_desc, TLSIOConfig io_config):
    xio = XIO()
    xio.create(io_desc._c_value, &io_config._c_value)
    return xio


cpdef xio_from_saslioconfig(SASLClientIOConfig io_config):
    cdef const  c_xio.IO_INTERFACE_DESCRIPTION* interface
    interface = c_sasl_mechanism.saslclientio_get_interface_description()
    if <void*>interface == NULL:
        raise ValueError("Failed to create SASL Client IO Interface description")
    xio = XIO()
    xio.create(interface, &io_config._c_value)
    return xio


cdef class XIO(StructBase):

    cdef c_xio.XIO_HANDLE _c_value

    def __cinit__(self):
        pass

    def __dealloc__(self):
        _logger.debug("Deallocating XIO")
        self.destroy()

    cdef _create(self):
        if <void*>self._c_value is NULL:
            self._memory_error()

    cpdef destroy(self):
        if <void*>self._c_value is not NULL:
            _logger.debug("Destroying XIO")
            c_xio.xio_destroy(self._c_value)
            self._c_value = <c_xio.XIO_HANDLE>NULL

    cdef wrap(self, c_xio.XIO_HANDLE value):
        self.destroy()
        self._c_value = value
        self._create()

    cdef create(self, c_xio.IO_INTERFACE_DESCRIPTION* io_desc, void *io_params):
        self.destroy()
        self._c_value = c_xio.xio_create(io_desc, io_params)
        self._create()

    cpdef set_option(self, const char* option_name, value):
        cdef const void* option_value
        option_value = <const void*>value
        if c_xio.xio_setoption(self._c_value, option_name, option_value) != 0:
            raise self._value_error("Failed to set option {}".format(option_name))

    cpdef set_certificates(self, bytes value):
        cdef char *certificate = value
        if c_xio.xio_setoption(self._c_value, b'TrustedCerts', <void*>certificate) != 0:
            raise self._value_error("Failed to set certificates")


cdef class IOInterfaceDescription(object):

    cdef c_xio.IO_INTERFACE_DESCRIPTION* _c_value

    def __cinit__(self):
        pass

    cdef wrap(self, c_xio.IO_INTERFACE_DESCRIPTION* value):
        self._c_value = value



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

import sys
from libc cimport stdint

cimport c_strings


cdef extern from "Python.h":

    ctypedef int Py_intptr_t
    ctypedef size_t Py_ssize_t


cdef extern from "azure_c_shared_utility/sastoken.h":
    bint SASToken_Validate(c_strings.STRING_HANDLE sasToken)
    c_strings.STRING_HANDLE SASToken_Create(c_strings.STRING_HANDLE key, c_strings.STRING_HANDLE scope, c_strings.STRING_HANDLE keyName, size_t expiry)
    c_strings.STRING_HANDLE SASToken_CreateString(const char* key, const char* scope, const char* keyName, size_t expiry)


cdef extern from "azure_c_shared_utility/buffer_.h":

    ctypedef struct BUFFER_HANDLE:
        pass

    BUFFER_HANDLE BUFFER_create(const unsigned char* source, size_t size);
    void BUFFER_delete(BUFFER_HANDLE handle)


cdef extern from "azure_c_shared_utility/map.h":

    cdef enum MAP_RESULT_TAG:
        MAP_OK,
        MAP_ERROR,
        MAP_INVALIDARG,
        MAP_KEYEXISTS,
        MAP_KEYNOTFOUND,
        MAP_FILTER_REJECT

    ctypedef struct MAP_HANDLE:
        pass

    ctypedef int (*MAP_FILTER_CALLBACK)(const char* mapProperty, const char* mapValue)

    MAP_RESULT_TAG Map_GetInternals(MAP_HANDLE handle, const char*const** keys, const char*const** values, size_t* count)
    MAP_HANDLE Map_Create(MAP_FILTER_CALLBACK mapFilterFunc)
    MAP_RESULT_TAG Map_Add(MAP_HANDLE handle, const char* key, const char* value)
    MAP_RESULT_TAG Map_AddOrUpdate(MAP_HANDLE handle, const char* key, const char* value)
    void Map_Destroy(MAP_HANDLE handle)


cdef extern from "azure_c_shared_utility/urlencode.h":

    c_strings.STRING_HANDLE URL_EncodeString(const char* textEncode)
    c_strings.STRING_HANDLE URL_Encode(c_strings.STRING_HANDLE input)


cdef extern from "azure_c_shared_utility/base64.h":

    c_strings.STRING_HANDLE Base64_Encoder(BUFFER_HANDLE input)
    c_strings.STRING_HANDLE Base64_Encode_Bytes(const unsigned char* source, size_t size)
    BUFFER_HANDLE Base64_Decoder(const char* source)


cdef extern from "azure_c_shared_utility/tickcounter.h":

    ctypedef struct TICK_COUNTER_HANDLE:
        pass

    # TODO: This is only for windows
    ctypedef stdint.uint_fast64_t tickcounter_ms_t  # Use 64-bit because of QueryPerformanceCounter call
    void tickcounter_destroy(TICK_COUNTER_HANDLE tick_counter)
    TICK_COUNTER_HANDLE tickcounter_create()
    int tickcounter_get_current_ms(TICK_COUNTER_HANDLE tick_counter, tickcounter_ms_t* current_ms)


cdef extern from "azure_c_shared_utility/doublylinkedlist.h":

    ctypedef struct DLIST_ENTRY_TAG:
        DLIST_ENTRY_TAG *Flink
        DLIST_ENTRY_TAG *Blink

    ctypedef DLIST_ENTRY_TAG DLIST_ENTRY
    ctypedef DLIST_ENTRY_TAG *PDLIST_ENTRY

    void DList_InitializeListHead(PDLIST_ENTRY listHead)
    int DList_IsListEmpty(const PDLIST_ENTRY listHead)
    void DList_InsertTailList(PDLIST_ENTRY listHead, PDLIST_ENTRY listEntry)
    void DList_InsertHeadList(PDLIST_ENTRY listHead, PDLIST_ENTRY listEntry)
    void DList_AppendTailList(PDLIST_ENTRY listHead, PDLIST_ENTRY ListToAppend)
    int DList_RemoveEntryList(PDLIST_ENTRY listEntry)
    PDLIST_ENTRY DList_RemoveHeadList(PDLIST_ENTRY listHead)


cdef extern from "azure_c_shared_utility/optionhandler.h":

    cdef enum OPTIONHANDLER_RESULT_TAG:
        OPTIONHANDLER_OK,
        OPTIONHANDLER_ERROR,
        OPTIONHANDLER_INVALIDARG

    ctypedef struct OPTIONHANDLER_HANDLE:
        pass

    ctypedef void* (*pfCloneOption)(const char* name, const void* value)
    ctypedef void (*pfDestroyOption)(const char* name, const void* value)
    ctypedef int (*pfSetOption)(void* handle, const char* name, const void* value)

    OPTIONHANDLER_HANDLE OptionHandler_Create(pfCloneOption cloneOption, pfDestroyOption destroyOption, pfSetOption setOption)
    OPTIONHANDLER_HANDLE OptionHandler_Clone(OPTIONHANDLER_HANDLE handler)
    OPTIONHANDLER_RESULT_TAG OptionHandler_AddOption(OPTIONHANDLER_HANDLE handle, const char* name, const void* value)
    OPTIONHANDLER_RESULT_TAG OptionHandler_FeedOptions(OPTIONHANDLER_HANDLE handle, void* destinationHandle)
    void OptionHandler_Destroy(OPTIONHANDLER_HANDLE handle)



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

# Python imports
import logging

# C imports
from libc cimport stdint
cimport c_session
cimport c_connection


_logger = logging.getLogger(__name__)


cpdef create_session(Connection connection, on_link_attached_context):
    session = cSession()
    session.create(connection._c_value, <c_session.ON_LINK_ATTACHED>on_link_attached, <void*>on_link_attached_context)
    return session


cdef class cSession(StructBase):

    _links = []
    cdef c_session.SESSION_HANDLE _c_value

    def __cinit__(self):
        pass

    def __dealloc__(self):
        _logger.debug("Deallocating cSession")
        self.destroy()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.destroy()

    @property
    def incoming_window(self):
        cdef stdint.uint32_t value
        if c_session.session_get_incoming_window(self._c_value, &value) != 0:
            self._value_error()
        return value

    @incoming_window.setter
    def incoming_window(self, stdint.uint32_t value):
        if c_session.session_set_incoming_window(self._c_value, value) != 0:
            self._value_error()

    @property
    def outgoing_window(self):
        cdef stdint.uint32_t value
        if c_session.session_get_outgoing_window(self._c_value, &value) != 0:
            self._value_error()
        return value

    @outgoing_window.setter
    def outgoing_window(self, stdint.uint32_t value):
        if c_session.session_set_outgoing_window(self._c_value, value) != 0:
            self._value_error()

    @property
    def handle_max(self):
        cdef c_amqp_definitions.handle value
        if c_session.session_get_handle_max(self._c_value, &value) != 0:
            self._value_error()
        return value

    @handle_max.setter
    def handle_max(self, c_amqp_definitions.handle value):
        if c_session.session_set_handle_max(self._c_value, value) != 0:
            self._value_error()

    cdef _create(self):
        if <void*>self._c_value is NULL:
            self._memory_error()

    cpdef destroy(self):
        if <void*>self._c_value is not NULL:
            _logger.debug("Destroying cSession")
            c_session.session_destroy(self._c_value)
            self._c_value = <c_session.SESSION_HANDLE>NULL

    cdef wrap(self, c_session.SESSION_HANDLE value):
        self.destroy()
        self._c_value = value
        self._create()

    cdef create(self, c_connection.CONNECTION_HANDLE connection, c_session.ON_LINK_ATTACHED on_link_attached, void* callback_context):
        self.destroy()
        self._c_value = c_session.session_create(connection, on_link_attached, callback_context)
        self._create()

    cpdef begin(self):
        if c_session.session_begin(self._c_value) != 0:
            self._value_error()

    cpdef end(self, const char* condition_value, const char* description):
        if c_session.session_end(self._c_value, condition_value, description) != 0:
            self._value_error()


#### Callback

cdef bint on_link_attached(
        void* context, c_session.LINK_ENDPOINT_HANDLE new_link_endpoint, const char* name,
        c_amqp_definitions.role role, c_amqpvalue.AMQP_VALUE source, c_amqpvalue.AMQP_VALUE target,
        c_amqp_definitions.fields properties):

    cdef c_amqp_definitions.SOURCE_HANDLE wrapped_source
    cdef c_amqp_definitions.TARGET_HANDLE wrapped_target
    cdef c_amqpvalue.AMQP_VALUE wrapped_props
    cdef c_amqpvalue.AMQP_VALUE cloned_source
    cdef c_amqpvalue.AMQP_VALUE cloned_target
    cdef stdint.uint32_t _value
    attach_properties = None

    context_obj = <object>context
    if <void*>context_obj == NULL:
        return True
    if <void*>properties != NULL:
        wrapped_props = c_amqpvalue.amqpvalue_clone(properties)
        attach_properties = value_factory(wrapped_props)

    if <void*>source == NULL or <void*>target == NULL:
        _logger.info("Link ATTACH frame missing source and/or target. DETACH pending.")
        context_obj._attach_received(None, None, attach_properties, "ATTACH frame Source and/or Target is NULL")
    else:
        try:
            cloned_source = c_amqpvalue.amqpvalue_clone(source)
            cloned_target = c_amqpvalue.amqpvalue_clone(target)
            if c_amqp_definitions.amqpvalue_get_source(cloned_source, &wrapped_source) != 0:
                context_obj._attach_received(None, None, attach_properties, "Unable to decode ATTACH frame source")
            elif c_amqp_definitions.amqpvalue_get_target(cloned_target, &wrapped_target) != 0:
                context_obj._attach_received(None, None, attach_properties, "Unable to decode ATTACH frame target")
            else:
                context_obj._attach_received(source_factory(wrapped_source), target_factory(wrapped_target), attach_properties, None)
        except Exception as e:
            _logger.info("Failed to process link ATTACH frame: %r", e)
    return True



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

import logging

import six
from uamqp import c_uamqp, compat, constants, utils

_logger = logging.getLogger(__name__)


class Address(object):
    """Represents an AMQP endpoint.

    :ivar address: The endpoint URL.
    :vartype address: str
    :ivar durable: Whether the endpoint is durable.
    :vartype: bool
    :ivar expiry_policy: The endpoint expiry policy
    :ivar timeout: The endpoint timeout in seconds.
    :vartype timeout: float
    :ivar dynamic: Whether the endpoint is dynamic.
    :vartype dynamic: bool
    :ivar distribution_mode: The endpoint distribution mode.
    :vartype distribution_mode: str
    :param address: An AMQP endpoint URL.
    :type address: str or bytes
    :param encoding: The encoding used if address is supplied
     as a str rather than bytes. Default is UTF-8.
    """

    def __init__(self, address, encoding='UTF-8'):
        address = address.encode(encoding) if isinstance(address, six.text_type) else address
        self.parsed_address = self._validate_address(address)
        self._encoding = encoding
        self._address = None
        addr = self.parsed_address.path
        if self.parsed_address.hostname:
            addr = self.parsed_address.hostname + addr
        if self.parsed_address.scheme:
            addr = self.parsed_address.scheme + b"://" + addr
        self._c_address = c_uamqp.string_value(addr)

    @classmethod
    def from_c_obj(cls, c_value, encoding='UTF-8'):
        address = c_value.address
        py_obj = cls(address, encoding=encoding)
        py_obj._address = c_value  # pylint: disable=protected-access
        return py_obj


    def __repr__(self):
        """Get the Address as a URL.

        :rtype: bytes
        """
        return self.parsed_address.geturl()

    def __str__(self):
        """Get the Address as a URL.

        :rtype: str
        """
        return self.parsed_address.geturl().decode(self._encoding)

    @property
    def hostname(self):
        return self.parsed_address.hostname.decode(self._encoding)

    @property
    def scheme(self):
        return self.parsed_address.scheme.decode(self._encoding)

    @property
    def username(self):
        if self.parsed_address.username:
            return self.parsed_address.username.decode(self._encoding)
        return None

    @property
    def password(self):
        if self.parsed_address.password:
            return self.parsed_address.password.decode(self._encoding)
        return None

    @property
    def address(self):
        return self._address.address.decode(self._encoding)

    @property
    def durable(self):
        return self._address.durable

    @durable.setter
    def durable(self, value):
        self._address.durable = value

    @property
    def expiry_policy(self):
        return self._address.expiry_policy

    @expiry_policy.setter
    def expiry_policy(self, value):
        self._address.expiry_policy = value

    @property
    def timeout(self):
        return self._address.timeout

    @timeout.setter
    def timeout(self, value):
        self._address.timeout = value

    @property
    def dynamic(self):
        return self._address.dynamic

    @dynamic.setter
    def dynamic(self, value):
        self._address.dynamic = value

    @property
    def distribution_mode(self):
        return self._address.distribution_mode.decode(self._encoding)

    @distribution_mode.setter
    def distribution_mode(self, value):
        mode = value.encode(self._encoding) if isinstance(value, six.text_type) else value
        self._address.distribution_mode = mode

    def _validate_address(self, address):
        """Confirm that supplied address is a valid URL and
        has an `amqp` or `amqps` scheme.

        :param address: The endpiont URL.
        :type address: str
        :rtype: ~urllib.parse.ParseResult
        """
        parsed = compat.urlparse(address)
        if not parsed.path:
            raise ValueError("Invalid {} address: {}".format(
                self.__class__.__name__, parsed))
        return parsed


class Source(Address):
    """Represents an AMQP Source endpoint.

    :ivar address: The endpoint URL.
    :vartype address: str
    :ivar durable: Whether the endpoint is durable.
    :vartype: bool
    :ivar expiry_policy: The endpoint expiry policy
    :ivar timeout: The endpoint timeout in seconds.
    :vartype timeout: float
    :ivar dynamic: Whether the endpoint is dynamic.
    :vartype dynamic: bool
    :ivar distribution_mode: The endpoint distribution mode.
    :vartype distribution_mode: str

    :param address: An AMQP endpoint URL.
    :type address: str or bytes
    :param encoding: The encoding used if address is supplied
     as a str rather than bytes. Default is UTF-8.
    """

    def __init__(self, address, encoding='UTF-8'):
        super(Source, self).__init__(address, encoding)
        self.filter_key = constants.STRING_FILTER
        self._address = c_uamqp.create_source()
        self._address.address = self._c_address

    def get_filter(self, name=constants.STRING_FILTER):
        """Get the filter on the source.

        :param name: The name of the filter. This will be encoded as
         an AMQP Symbol. By default this is set to b'apache.org:selector-filter:string'.
        :type name: bytes
        """
        try:
            filter_key = c_uamqp.symbol_value(name)
            return self._address.filter_set[filter_key].value
        except (TypeError, KeyError):
            return None

    def set_filter(self, value, name=constants.STRING_FILTER, descriptor=constants.STRING_FILTER):
        """Set a filter on the endpoint. Only one filter
        can be applied to an endpoint.

        :param value: The filter to apply to the endpoint. Set to None for a NULL filter.
        :type value: bytes or str or None
        :param name: The name of the filter. This will be encoded as
         an AMQP Symbol. By default this is set to b'apache.org:selector-filter:string'.
        :type name: bytes
        :param descriptor: The descriptor used if the filter is to be encoded as a described value.
         This will be encoded as an AMQP Symbol. By default this is set to b'apache.org:selector-filter:string'.
         Set to None if the filter should not be encoded as a described value.
        :type descriptor: bytes or None
        """
        value = value.encode(self._encoding) if isinstance(value, six.text_type) else value
        filter_set = c_uamqp.dict_value()
        filter_key = c_uamqp.symbol_value(name)
        filter_value = utils.data_factory(value, encoding=self._encoding)
        if value is not None and descriptor is not None:
            descriptor = c_uamqp.symbol_value(descriptor)
            filter_value = c_uamqp.described_value(descriptor, filter_value)

        filter_set[filter_key] = filter_value
        self._address.filter_set = filter_set


class Target(Address):
    """Represents an AMQP Target endpoint.

    :ivar address: The endpoint URL.
    :vartype address: str
    :ivar durable: Whether the endpoint is durable.
    :vartype: bool
    :ivar expiry_policy: The endpoint expiry policy
    :ivar timeout: The endpoint timeout in seconds.
    :vartype timeout: float
    :ivar dynamic: Whether the endpoint is dynamic.
    :vartype dynamic: bool
    :ivar distribution_mode: The endpoint distribution mode.
    :vartype distribution_mode: str

    :param address: An AMQP endpoint URL.
    :type address: str or bytes
    :param encoding: The encoding used if address is supplied
     as a str rather than bytes. Default is UTF-8.
    """

    def __init__(self, address, encoding='UTF-8'):
        super(Target, self).__init__(address, encoding)
        self._address = c_uamqp.create_target()
        self._address.address = self._c_address



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

from libc cimport stdint

cimport c_xio


cdef extern from "azure_c_shared_utility/wsio.h":

    ctypedef struct WSIO_CONFIG_TAG:
        const char* hostname
        int port
        const char* resource_name
        const char* protocol
        const c_xio.IO_INTERFACE_DESCRIPTION* underlying_io_interface
        void* underlying_io_parameters

    ctypedef WSIO_CONFIG_TAG WSIO_CONFIG

    const c_xio.IO_INTERFACE_DESCRIPTION* wsio_get_interface_description()



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

import base64
import calendar
import time
import uuid
import logging
from datetime import timedelta, datetime

import six
from uamqp import c_uamqp

logger = logging.getLogger(__name__)


def get_running_loop():
    try:
        import asyncio  # pylint: disable=import-error
        return asyncio.get_running_loop()
    except AttributeError:  # 3.5 / 3.6
        loop = None
        try:
            loop = asyncio._get_running_loop()  # pylint: disable=protected-access
        except AttributeError:
            logger.warning('This version of Python is deprecated, please upgrade to >= v3.5.3')
        if loop is None:
            logger.warning('No running event loop')
            loop = asyncio.get_event_loop()
        return loop
    except RuntimeError:
        # For backwards compatibility, create new event loop
        logger.warning('No running event loop')
        return asyncio.get_event_loop()


def parse_connection_string(connect_str):
    """Parse a connection string such as those provided by the Azure portal.
    Connection string should be formatted like: `Key=Value;Key=Value;Key=Value`.
    The connection string will be parsed into a dictionary.

    :param connect_str: The connection string.
    :type connect_str: str
    :rtype: dict[str, str]
    """
    connect_info = {}
    fields = connect_str.split(';')
    for field in fields:
        key, value = field.split('=', 1)
        connect_info[key] = value
    return connect_info


def create_sas_token(key_name, shared_access_key, scope, expiry=timedelta(hours=1)):
    """Create a SAS token.

    :param key_name: The username/key name/policy name for the token.
    :type key_name: bytes
    :param shared_access_key: The shared access key to generate the token from.
    :type shared_access_key: bytes
    :param scope: The token permissions scope.
    :type scope: bytes
    :param expiry: The lifetime of the generated token. Default is 1 hour.
    :type expiry: ~datetime.timedelta
    :rtype: bytes
    """
    shared_access_key = base64.b64encode(shared_access_key)
    abs_expiry = int(time.time()) + expiry.seconds
    return c_uamqp.create_sas_token(shared_access_key, scope, key_name, abs_expiry)


def _convert_py_number(value):
    """Convert a Python integer value into equivalent C object.
    Will attempt to use the smallest possible conversion, starting with int, then long
    then double.
    """
    try:
        return c_uamqp.int_value(value)
    except OverflowError:
        pass
    try:
        return c_uamqp.long_value(value)
    except OverflowError:
        pass
    return c_uamqp.double_value(value)


def data_factory(value, encoding='UTF-8'):
    """Wrap a Python type in the equivalent C AMQP type.
    If the Python type has already been wrapped in a ~uamqp.types.AMQPType
    object - then this will be used to select the appropriate C type.
    - bool => c_uamqp.BoolValue
    - int => c_uamqp.IntValue, LongValue, DoubleValue
    - str => c_uamqp.StringValue
    - bytes => c_uamqp.BinaryValue
    - list/set/tuple => c_uamqp.ListValue
    - dict => c_uamqp.DictValue (AMQP map)
    - float => c_uamqp.DoubleValue
    - uuid.UUID => c_uamqp.UUIDValue

    :param value: The value to wrap.
    :type value: ~uamqp.types.AMQPType
    :rtype: uamqp.c_uamqp.AMQPValue
    """
    result = None
    if value is None:
        result = c_uamqp.null_value()
    elif hasattr(value, 'c_data'):
        result = value.c_data
    elif isinstance(value, c_uamqp.AMQPValue):
        result = value
    elif isinstance(value, bool):
        result = c_uamqp.bool_value(value)
    elif isinstance(value, six.text_type):
        result = c_uamqp.string_value(value.encode(encoding))
    elif isinstance(value, six.binary_type):
        result = c_uamqp.string_value(value)
    elif isinstance(value, uuid.UUID):
        result = c_uamqp.uuid_value(value)
    elif isinstance(value, bytearray):
        result = c_uamqp.binary_value(value)
    elif isinstance(value, six.integer_types):
        result = _convert_py_number(value)
    elif isinstance(value, float):
        result = c_uamqp.double_value(value)
    elif isinstance(value, dict):
        wrapped_dict = c_uamqp.dict_value()
        for key, item in value.items():
            wrapped_dict[data_factory(key, encoding=encoding)] = data_factory(item, encoding=encoding)
        result = wrapped_dict
    elif isinstance(value, (list, set, tuple)):
        wrapped_list = c_uamqp.list_value()
        wrapped_list.size = len(value)
        for index, item in enumerate(value):
            wrapped_list[index] = data_factory(item, encoding=encoding)
        result = wrapped_list
    elif isinstance(value, datetime):
        timestamp = int((calendar.timegm(value.utctimetuple()) * 1000) + (value.microsecond/1000))
        result = c_uamqp.timestamp_value(timestamp)
    return result



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

# Python imports
import logging

# C imports
from libc cimport stdint

cimport c_link
cimport c_session
cimport c_amqp_definitions
cimport c_amqpvalue


_logger = logging.getLogger(__name__)


cpdef create_link(cSession session, const char* name, bint role, AMQPValue source, AMQPValue target):
    new_link = cLink()
    new_link.create(<c_session.SESSION_HANDLE>session._c_value, name, <c_amqp_definitions.role>role, <c_amqpvalue.AMQP_VALUE>source._c_value, <c_amqpvalue.AMQP_VALUE>target._c_value)
    return new_link


cdef class cLink(StructBase):

    cdef c_link.LINK_HANDLE _c_value
    cdef c_link.ON_LINK_DETACH_EVENT_SUBSCRIPTION_HANDLE _detach_event

    def __cinit__(self):
        pass

    def __dealloc__(self):
        _logger.debug("Deallocating cLink")
        self.destroy()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.destroy()

    cdef _create(self):
        if <void*>self._c_value is NULL:
            self._memory_error()

    cpdef destroy(self):
        if <void*>self._c_value is not NULL:
            _logger.debug("Destroying cLink")
            c_link.link_destroy(self._c_value)
            self._c_value = <c_link.LINK_HANDLE>NULL

    cdef wrap(self, c_link.LINK_HANDLE value):
        self.destroy()
        self._c_value = value
        self._create()

    cdef create(self, c_session.SESSION_HANDLE session, const char* name, c_amqp_definitions.role role, c_amqpvalue.AMQP_VALUE source, c_amqpvalue.AMQP_VALUE target):
        self.destroy()
        self._c_value = c_link.link_create(session, name, role, source, target)
        self._create()

    cpdef subscribe_to_detach_event(self, on_detch_received):
        self._detach_event = c_link.link_subscribe_on_link_detach_received(
            self._c_value,
            <c_link.ON_LINK_DETACH_RECEIVED>on_link_detach_received,
            <void*>on_detch_received)
        if <void*>self._detach_event == NULL:
            self._value_error("Unable to register DETACH event handler.")

    cpdef unsubscribe_to_detach_event(self):
        c_link.link_unsubscribe_on_link_detach_received(self._detach_event)

    @property
    def send_settle_mode(self):
        cdef c_amqp_definitions.sender_settle_mode snd_settle_mode
        if  c_link.link_get_snd_settle_mode(self._c_value, &snd_settle_mode) != 0:
            self._value_error()
        return <stdint.uint8_t>snd_settle_mode

    @send_settle_mode.setter
    def send_settle_mode(self, stdint.uint8_t value):
        if c_link.link_set_snd_settle_mode(self._c_value, <c_amqp_definitions.sender_settle_mode>value) != 0:
            self._value_error()

    @property
    def receive_settle_mode(self):
        cdef c_amqp_definitions.receiver_settle_mode rcv_settle_mode
        if  c_link.link_get_rcv_settle_mode(self._c_value, &rcv_settle_mode) != 0:
            self._value_error()
        return <stdint.uint8_t>rcv_settle_mode

    @receive_settle_mode.setter
    def receive_settle_mode(self, stdint.uint8_t value):
        if c_link.link_set_rcv_settle_mode(self._c_value, <c_amqp_definitions.receiver_settle_mode>value) != 0:
            self._value_error()

    @property
    def max_message_size(self):
        cdef stdint.uint64_t value
        if c_link.link_get_max_message_size(self._c_value, &value) != 0:
            self._value_error()
        return value

    @max_message_size.setter
    def max_message_size(self, stdint.uint64_t value):
        if c_link.link_set_max_message_size(self._c_value, value) != 0:
            self._value_error()

    @property
    def initial_delivery_count(self):
        cdef c_amqp_definitions.sequence_no value
        if c_link.link_get_initial_delivery_count(self._c_value, &value) != 0:
            self._value_error()
        return value

    @initial_delivery_count.setter
    def initial_delivery_count(self, c_amqp_definitions.sequence_no value):
        if c_link.link_set_initial_delivery_count(self._c_value, value) != 0:
            self._value_error()

    @property
    def peer_max_message_size(self):
        cdef stdint.uint64_t value
        if c_link.link_get_peer_max_message_size(self._c_value, &value) != 0:
            self._value_error()
        return value

    @property
    def name(self):
        cdef const char* value
        if c_link.link_get_name(self._c_value, &value) != 0:
            self._value_error()
        return value

    cpdef set_prefetch_count(self, stdint.uint32_t prefetch):
        if c_link.link_set_max_link_credit(self._c_value, prefetch) != 0:
            self._value_error("Unable to set link credit.")

    cpdef set_attach_properties(self, AMQPValue properties):
        if c_link.link_set_attach_properties(self._c_value, <c_amqp_definitions.fields>properties._c_value) != 0:
            self._value_error("Unable to set link attach properties.")


#### Callback

cdef void on_link_detach_received(void* context, c_amqp_definitions.ERROR_HANDLE error):
    cdef c_amqp_definitions.ERROR_HANDLE cloned
    context_obj = <object>context
    if <void*> error != NULL:
        cloned = c_amqp_definitions.error_clone(error)
        wrapped_error = error_factory(cloned)
    else:
        wrapped_error = None
    if hasattr(context_obj, '_detach_received'):
        context_obj._detach_received(wrapped_error)



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

# Python imports
import logging

# C imports
cimport c_strings
cimport c_platform
cimport c_xio
cimport c_tlsio


_logger = logging.getLogger(__name__)


cpdef platform_init():
    if c_platform.platform_init() != 0:
        raise ValueError("Failed to initialize platform.")


cpdef platform_deinit():
    _logger.debug("Deinitializing platform")
    c_platform.platform_deinit()


cpdef get_info():
    cdef c_strings.STRING_HANDLE str_info
    str_info = c_platform.platform_get_platform_info()
    info = AMQPString()
    info.wrap(str_info)
    return info


cpdef get_default_tlsio():
    cdef const c_xio.IO_INTERFACE_DESCRIPTION* io_desc
    io_desc = c_platform.platform_get_default_tlsio()
    if <void*>io_desc == NULL:
        raise ValueError("Failed to create tlsio description.")

    interface = IOInterfaceDescription()
    interface.wrap(io_desc)
    return interface



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

from libc cimport stdint

cimport c_amqpvalue
cimport c_amqp_definitions


cdef extern from "azure_uamqp_c/messaging.h":

    c_amqpvalue.AMQP_VALUE messaging_create_source(const char* address)
    c_amqpvalue.AMQP_VALUE messaging_create_target(const char* address)

    c_amqpvalue.AMQP_VALUE messaging_delivery_received(stdint.uint32_t section_number, stdint.uint64_t section_offset) nogil
    c_amqpvalue.AMQP_VALUE messaging_delivery_accepted() nogil
    c_amqpvalue.AMQP_VALUE messaging_delivery_rejected(const char* error_condition, const char* error_description) nogil
    c_amqpvalue.AMQP_VALUE messaging_delivery_released() nogil
    c_amqpvalue.AMQP_VALUE messaging_delivery_modified(bint delivery_failed, bint undeliverable_here, c_amqp_definitions.fields message_annotations) nogil


cdef extern from "azure_uamqp_c/message.h":

    cdef enum MESSAGE_BODY_TYPE_TAG:
        MESSAGE_BODY_TYPE_NONE,
        MESSAGE_BODY_TYPE_DATA,
        MESSAGE_BODY_TYPE_SEQUENCE,
        MESSAGE_BODY_TYPE_VALUE,

    ctypedef struct MESSAGE_HANDLE:
        pass

    cdef struct BINARY_DATA_TAG:
        const unsigned char* bytes;
        size_t length;

    ctypedef BINARY_DATA_TAG BINARY_DATA

    MESSAGE_HANDLE message_create()
    MESSAGE_HANDLE message_clone(MESSAGE_HANDLE source_message)
    void message_destroy(MESSAGE_HANDLE message)
    int message_set_header(MESSAGE_HANDLE message, c_amqp_definitions.HEADER_HANDLE message_header)
    int message_get_header(MESSAGE_HANDLE message, c_amqp_definitions.HEADER_HANDLE* message_header)
    int message_set_delivery_annotations(MESSAGE_HANDLE message, c_amqp_definitions.delivery_annotations annotations)
    int message_get_delivery_annotations(MESSAGE_HANDLE message, c_amqp_definitions.delivery_annotations* annotations)
    int message_set_message_annotations(MESSAGE_HANDLE message, c_amqp_definitions.message_annotations annotations)
    int message_get_message_annotations(MESSAGE_HANDLE message, c_amqp_definitions.message_annotations* annotations)
    int message_set_properties(MESSAGE_HANDLE message, c_amqp_definitions.PROPERTIES_HANDLE properties)
    int message_get_properties(MESSAGE_HANDLE message, c_amqp_definitions.PROPERTIES_HANDLE* properties)
    int message_set_application_properties(MESSAGE_HANDLE message, c_amqpvalue.AMQP_VALUE application_properties)
    int message_get_application_properties(MESSAGE_HANDLE message, c_amqpvalue.AMQP_VALUE* application_properties)
    int message_set_footer(MESSAGE_HANDLE message, c_amqp_definitions.annotations footer)
    int message_get_footer(MESSAGE_HANDLE message, c_amqp_definitions.annotations* footer)
    
    int message_add_body_amqp_data(MESSAGE_HANDLE message, BINARY_DATA amqp_data)
    int message_get_body_amqp_data_in_place(MESSAGE_HANDLE message, size_t index, BINARY_DATA* amqp_data)
    int message_get_body_amqp_data_count(MESSAGE_HANDLE message, size_t* count)

    int message_set_body_amqp_value(MESSAGE_HANDLE message, c_amqpvalue.AMQP_VALUE body_amqp_value)
    int message_get_body_amqp_value_in_place(MESSAGE_HANDLE message, c_amqpvalue.AMQP_VALUE* body_amqp_value)

    int message_add_body_amqp_sequence(MESSAGE_HANDLE message, c_amqpvalue.AMQP_VALUE sequence)
    int message_get_body_amqp_sequence_in_place(MESSAGE_HANDLE message, size_t index, c_amqpvalue.AMQP_VALUE* sequence)
    int message_get_body_amqp_sequence_count(MESSAGE_HANDLE message, size_t* count)

    int message_get_body_type(MESSAGE_HANDLE message, MESSAGE_BODY_TYPE_TAG* body_type)
    int message_set_message_format(MESSAGE_HANDLE message, stdint.uint32_t message_format)
    int message_get_message_format(MESSAGE_HANDLE message, stdint.uint32_t* message_format)





#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

# pylint: disable=too-many-lines

import logging

import six
from uamqp import c_uamqp, constants, errors, utils

_logger = logging.getLogger(__name__)


class Message(object):
    """An AMQP message.

    When sending, depending on the nature of the data,
    different body encoding will be used. If the data is str or bytes,
    a single part DataBody will be sent. If the data is a list of str/bytes,
    a multipart DataBody will be sent. Any other type of list or any other
    type of data will be sent as a ValueBody.
    An empty payload will also be sent as a ValueBody.

    :ivar on_send_complete: A custom callback to be run on completion of
     the send operation of this message. The callback must take two parameters,
     a result (of type `MessageSendResult`) and an error (of type
     Exception). The error parameter may be None if no error ocurred or the error
     information was undetermined.
    :vartype on_send_complete: callable[~uamqp.constants.MessageSendResult, Exception]

    :param body: The data to send in the message.
    :type body: Any Python data type.
    :param properties: Properties to add to the message.
    :type properties: ~uamqp.message.MessageProperties
    :param application_properties: Service specific application properties.
    :type application_properties: dict
    :param annotations: Service specific message annotations. Keys in the dictionary
     must be `types.AMQPSymbol` or `types.AMQPuLong`.
    :type annotations: dict
    :param header: The message header.
    :type header: ~uamqp.message.MessageHeader
    :param msg_format: A custom message format. Default is 0.
    :type msg_format: int
    :param message: Internal only. This is used to wrap an existing message
     that has been received from an AMQP service. If specified, all other
     parameters will be ignored.
    :type message: uamqp.c_uamqp.cMessage
    :param settler: Internal only. This is used when wrapping an existing message
     that has been received from an AMQP service. Should only be specified together
     with `message` and is to settle the message.
    :type settler: callable[~uamqp.errors.MessageResponse]
    :param delivery_no: Internal only. This is used when wrapping an existing message
     that has been received from an AMQP service. Should only be specified together
     with `message` and specifies the messages client delivery number.
    :param encoding: The encoding to use for parameters supplied as strings.
     Default is 'UTF-8'
    :type encoding: str
    """

    def __init__(self,
                 body=None,
                 properties=None,
                 application_properties=None,
                 annotations=None,
                 header=None,
                 msg_format=None,
                 message=None,
                 settler=None,
                 delivery_no=None,
                 encoding='UTF-8'):
        self.state = constants.MessageState.WaitingToBeSent
        self.idle_time = 0
        self.retries = 0
        self._response = None
        self._settler = None
        self._encoding = encoding
        self.delivery_no = delivery_no
        self.on_send_complete = None
        self.properties = None
        self.application_properties = None
        self.annotations = None
        self.header = None
        self.footer = None
        self.delivery_annotations = None

        if message:
            if settler:
                self.state = constants.MessageState.ReceivedUnsettled
                self._response = None
            else:
                self.state = constants.MessageState.ReceivedSettled
                self._response = errors.MessageAlreadySettled()
            self._settler = settler
            self._parse_message(message)
        else:
            self._message = c_uamqp.create_message()
            if isinstance(body, (six.text_type, six.binary_type)):
                self._body = DataBody(self._message)
                self._body.append(body)
            elif isinstance(body, list) and all([isinstance(b, (six.text_type, six.binary_type)) for b in body]):
                self._body = DataBody(self._message)
                for value in body:
                    self._body.append(value)
            else:
                self._body = ValueBody(self._message)
                self._body.set(body)
            if msg_format:
                self._message.message_format = msg_format
            self.properties = properties
            self.application_properties = application_properties
            self.annotations = annotations
            self.header = header

    @classmethod
    def decode_from_bytes(cls, data):
        """Decode an AMQP message from a bytearray.
        The returned message will not have a delivery context and
        therefore will be considered to be in an "already settled" state.

        :param data: The AMQP wire-encoded bytes to decode.
        :type data: bytes or bytearray
        """
        decoded_message = c_uamqp.decode_message(len(data), data)
        return cls(message=decoded_message)

    def __str__(self):
        if not self._message:
            return ""
        return str(self._body)

    def _parse_message(self, message):
        """Parse a message received from an AMQP service.

        :param message: The received C message.
        :type message: uamqp.c_uamqp.cMessage
        """
        _logger.debug("Parsing received message %r.", self.delivery_no)
        self._message = message
        body_type = message.body_type
        if body_type == c_uamqp.MessageBodyType.NoneType:
            self._body = None
        elif body_type == c_uamqp.MessageBodyType.DataType:
            self._body = DataBody(self._message)
        elif body_type == c_uamqp.MessageBodyType.SequenceType:
            raise TypeError("Message body type Sequence not supported.")
        else:
            self._body = ValueBody(self._message)
        _props = self._message.properties
        if _props:
            _logger.debug("Parsing received message properties %r.", self.delivery_no)
            self.properties = MessageProperties(properties=_props, encoding=self._encoding)
        _header = self._message.header
        if _header:
            _logger.debug("Parsing received message header %r.", self.delivery_no)
            self.header = MessageHeader(header=_header)
        _footer = self._message.footer
        if _footer:
            _logger.debug("Parsing received message footer %r.", self.delivery_no)
            self.footer = _footer.map
        _app_props = self._message.application_properties
        if _app_props:
            _logger.debug("Parsing received message application properties %r.", self.delivery_no)
            self.application_properties = _app_props.map
        _ann = self._message.message_annotations
        if _ann:
            _logger.debug("Parsing received message annotations %r.", self.delivery_no)
            self.annotations = _ann.map
        _delivery_ann = self._message.delivery_annotations
        if _delivery_ann:
            _logger.debug("Parsing received message delivery annotations %r.", self.delivery_no)
            self.delivery_annotations = _delivery_ann.map

    def _can_settle_message(self):
        if self.state not in constants.RECEIVE_STATES:
            raise TypeError("Only received messages can be settled.")
        if self.settled:
            return False
        return True

    def _populate_message_attributes(self, c_message):
        if self.properties:
            c_message.properties = self.properties.get_properties_obj()
        if self.application_properties:
            if not isinstance(self.application_properties, dict):
                raise TypeError("Application properties must be a dictionary.")
            amqp_props = utils.data_factory(self.application_properties, encoding=self._encoding)
            c_message.application_properties = amqp_props
        if self.annotations:
            if not isinstance(self.annotations, dict):
                raise TypeError("Message annotations must be a dictionary.")
            ann_props = c_uamqp.create_message_annotations(
                utils.data_factory(self.annotations, encoding=self._encoding))
            c_message.message_annotations = ann_props
        if self.header:
            c_message.header = self.header.get_header_obj()

    @property
    def settled(self):
        """Whether the message transaction for this message has been completed.
        If this message is to be sent, the message will be `settled=True` once a
        disposition has been received from the service.
        If this message has been received, the message will be `settled=True` once
        a disposition has been sent to the service.

        :rtype: bool
        """
        if self._response:
            return True
        return False

    def get_message_encoded_size(self):
        """Pre-emptively get the size of the message once it has been encoded
        to go over the wire so we can raise an error if the message will be
        rejected for being to large.

        This method is not available for messages that have been received.

        :rtype: int
        """
        if not self._message:
            raise ValueError("No message data to encode.")
        cloned_data = self._message.clone()
        self._populate_message_attributes(cloned_data)
        encoded_data = []
        return c_uamqp.get_encoded_message_size(cloned_data, encoded_data)

    def encode_message(self):
        """Encode message to AMQP wire-encoded bytearray.

        :rtype: bytearray
        """
        if not self._message:
            raise ValueError("No message data to encode.")
        cloned_data = self._message.clone()
        self._populate_message_attributes(cloned_data)
        encoded_data = []
        c_uamqp.get_encoded_message_size(cloned_data, encoded_data)
        return b"".join(encoded_data)

    def get_data(self):
        """Get the body data of the message. The format may vary depending
        on the body type.

        :rtype: generator
        """
        if not self._message or not self._body:
            return None
        return self._body.data

    def gather(self):
        """Return all the messages represented by this object.
        This will always be a list of a single message.

        :rtype: list[~uamqp.message.Message]
        """
        if self.state in constants.RECEIVE_STATES:
            raise TypeError("Only new messages can be gathered.")
        if not self._message:
            raise ValueError("Message data already consumed.")
        try:
            raise self._response
        except TypeError:
            pass
        return [self]

    def get_message(self):
        """Get the underlying C message from this object.

        :rtype: uamqp.c_uamqp.cMessage
        """
        if not self._message:
            return None
        self._populate_message_attributes(self._message)
        return self._message

    def accept(self):
        """Send a response disposition to the service to indicate that
        a received message has been accepted. If the client is running in PeekLock
        mode, the service will wait on this disposition. Otherwise it will
        be ignored. Returns `True` is message was accepted, or `False` if the message
        was already settled.

        :rtype: bool
        :raises: TypeError if the message is being sent rather than received.
        """
        if self._can_settle_message():
            self._response = errors.MessageAccepted()
            self._settler(self._response)
            self.state = constants.MessageState.ReceivedSettled
            return True
        return False

    def reject(self, condition=None, description=None):
        """Send a response disposition to the service to indicate that
        a received message has been rejected. If the client is running in PeekLock
        mode, the service will wait on this disposition. Otherwise it will
        be ignored. A rejected message will increment the messages delivery count.
        Returns `True` is message was rejected, or `False` if the message
        was already settled.

        :param condition: The AMQP rejection code. By default this is `amqp:internal-error`.
        :type condition: bytes or str
        :param description: A description/reason to accompany the rejection.
        :type description: bytes or str
        :rtype: bool
        :raises: TypeError if the message is being sent rather than received.
        """
        if self._can_settle_message():
            self._response = errors.MessageRejected(
                condition=condition,
                description=description,
                encoding=self._encoding)
            self._settler(self._response)
            self.state = constants.MessageState.ReceivedSettled
            return True
        return False

    def release(self):
        """Send a response disposition to the service to indicate that
        a received message has been released. If the client is running in PeekLock
        mode, the service will wait on this disposition. Otherwise it will
        be ignored. A released message will not incremenet the messages
        delivery count. Returns `True` is message was released, or `False` if the message
        was already settled.

        :rtype: bool
        :raises: TypeError if the message is being sent rather than received.
        """
        if self._can_settle_message():
            self._response = errors.MessageReleased()
            self._settler(self._response)
            self.state = constants.MessageState.ReceivedSettled
            return True
        return False

    def modify(self, failed, deliverable, annotations=None):
        """Send a response disposition to the service to indicate that
        a received message has been modified. If the client is running in PeekLock
        mode, the service will wait on this disposition. Otherwise it will
        be ignored. Returns `True` is message was modified, or `False` if the message
        was already settled.

        :param failed: Whether this delivery of this message failed. This does not
         indicate whether subsequence deliveries of this message would also fail.
        :type failed: bool
        :param deliverable: Whether this message will be deliverable to this client
         on subsequent deliveries - i.e. whether delivery is retryable.
        :type deliverable: bool
        :param annotations: Annotations to attach to response.
        :type annotations: dict
        :rtype: bool
        :raises: TypeError if the message is being sent rather than received.
        """
        if self._can_settle_message():
            self._response = errors.MessageModified(
                failed,
                deliverable,
                annotations=annotations,
                encoding=self._encoding)
            self._settler(self._response)
            self.state = constants.MessageState.ReceivedSettled
            return True
        return False


class BatchMessage(Message):
    """A Batched AMQP message.

    This batch message encodes multiple message bodies into a single message
    to increase through-put over the wire. It requires server-side support
    to unpackage the batched messages and so will not be universally supported.

    :ivar on_send_complete: A custom callback to be run on completion of
     the send operation of this message. The callback must take two parameters,
     a result (of type ~uamqp.constants.MessageSendResult) and an error (of type
     Exception). The error parameter may be None if no error ocurred or the error
     information was undetermined.
    :vartype on_send_complete: callable[~uamqp.constants.MessageSendResult, Exception]
    :ivar batch_format: The is the specific message format to inform the service the
     the body should be interpreted as multiple messages. The value is 0x80013700.
    :vartype batch_format: int
    :ivar max_message_length: The maximum data size in bytes to allow in a single message.
     By default this is 256kb. If sending a single batch message, an error will be raised
     if the supplied data exceeds this maximum. If sending multiple batch messages, this
     value will be used to divide the supplied data between messages.
    :vartype max_message_length: int

    :param data: An iterable source of data, where each value will be considered the
     body of a single message in the batch.
    :type data: iterable
    :param properties: Properties to add to the message. If multiple messages are created
     these properties will be applied to each message.
    :type properties: ~uamqp.message.MessageProperties
    :param application_properties: Service specific application properties. If multiple messages
     are created these properties will be applied to each message.
    :type application_properties: dict
    :param annotations: Service specific message annotations. If multiple messages are created
     these properties will be applied to each message. Keys in the dictionary
     must be `types.AMQPSymbol` or `types.AMQPuLong`.
    :type annotations: dict
    :param header: The message header. This header will be applied to each message in the batch.
    :type header: ~uamqp.message.MessageHeader
    :param multi_messages: Whether to send the supplied data across multiple messages. If set to
     `False`, all the data will be sent in a single message, and an error raised if the message
     is too large. If set to `True`, the data will automatically be divided across multiple messages
     of an appropriate size. The default is `False`.
    :type multi_messages: bool
    :param encoding: The encoding to use for parameters supplied as strings.
     Default is 'UTF-8'
    :type encoding: str
    :raises: ValueError if data is sent in a single message and that message exceeds the max size.
    """

    batch_format = 0x80013700
    max_message_length = constants.MAX_MESSAGE_LENGTH_BYTES
    size_offset = 0

    def __init__(self,
                 data=None,
                 properties=None,
                 application_properties=None,
                 annotations=None,
                 header=None,
                 multi_messages=False,
                 encoding='UTF-8'):
        # pylint: disable=super-init-not-called
        self._multi_messages = multi_messages
        self._body_gen = data
        self._encoding = encoding
        self.on_send_complete = None
        self.properties = properties
        self.application_properties = application_properties
        self.annotations = annotations
        self.header = header

    def _create_batch_message(self):
        """Create a ~uamqp.message.Message for a value supplied by the data
        generator. Applies all properties and annotations to the message.

        :rtype: ~uamqp.message.Message
        """
        return Message(body=[],
                       properties=self.properties,
                       annotations=self.annotations,
                       msg_format=self.batch_format,
                       header=self.header,
                       encoding=self._encoding)

    def _multi_message_generator(self):
        """Generate multiple ~uamqp.message.Message objects from a single data
        stream that in total may exceed the maximum individual message size.
        Data will be continuously added to a single message until that message
        reaches a max allowable size, at which point it will be yielded and
        a new message will be started.

        :rtype: generator[~uamqp.message.Message]
        """
        unappended_message_bytes = None
        while True:
            new_message = self._create_batch_message()
            message_size = new_message.get_message_encoded_size() + self.size_offset
            body_size = 0
            if unappended_message_bytes:
                new_message._body.append(unappended_message_bytes)  # pylint: disable=protected-access
                body_size += len(unappended_message_bytes)
            try:
                for data in self._body_gen:
                    message_bytes = None
                    try:
                        if not data.application_properties and self.application_properties:  # Message-like object
                            data.application_properties = self.application_properties
                        message_bytes = data.encode_message()
                    except AttributeError:  # raw data
                        wrap_message = Message(body=data, application_properties=self.application_properties)
                        message_bytes = wrap_message.encode_message()
                    body_size += len(message_bytes)
                    if (body_size + message_size) > self.max_message_length:
                        new_message.on_send_complete = self.on_send_complete
                        unappended_message_bytes = message_bytes
                        yield new_message
                        raise StopIteration()
                    new_message._body.append(message_bytes)  # pylint: disable=protected-access
            except StopIteration:
                _logger.debug("Sent partial message.")
                continue
            else:
                new_message.on_send_complete = self.on_send_complete
                yield new_message
                _logger.debug("Sent all batched data.")
                break

    def gather(self):
        """Return all the messages represented by this object. This will convert
        the batch data into individual Message objects, which may be one
        or more if multi_messages is set to `True`.

        :rtype: list[~uamqp.message.Message]
        """
        if self._multi_messages:
            return self._multi_message_generator()

        new_message = self._create_batch_message()
        message_size = new_message.get_message_encoded_size() + self.size_offset
        body_size = 0

        for data in self._body_gen:
            message_bytes = None
            try:
                if not data.application_properties and self.application_properties:  # Message-like object
                    data.application_properties = self.application_properties
                message_bytes = data.encode_message()
            except AttributeError:  # raw data
                wrap_message = Message(body=data, application_properties=self.application_properties)
                message_bytes = wrap_message.encode_message()
            body_size += len(message_bytes)
            if (body_size + message_size) > self.max_message_length:
                raise errors.MessageContentTooLarge()
            new_message._body.append(message_bytes)  # pylint: disable=protected-access
        new_message.on_send_complete = self.on_send_complete
        return [new_message]


class MessageProperties(object):
    """Message properties.
    The properties that are actually used will depend on the service implementation.
    Not all received messages will have all properties, and not all properties
    will be utilized on a sent message.

    :ivar message_id: Message-id, if set, uniquely identifies a message within the message system.
     The message producer is usually responsible for setting the message-id in such a way that it
     is assured to be globally unique. A broker MAY discard a message as a duplicate if the value
     of the message-id matches that of a previously received message sent to the same node.
    :vartype message_id: str or bytes or uuid.UUID or ~uamqp.types.AMQPType
    :ivar user_id: The identity of the user responsible for producing the message. The client sets
     this value, and it MAY be authenticated by intermediaries.
    :vartype user_id: str or bytes
    :ivar to: The to field identifies the node that is the intended destination of the message.
     On any given transfer this might not be the node at the receiving end of the link.
    :vartype to: str or bytes
    :ivar subject:
    :vartype subject:
    :ivar reply_to:
    :vartype reply_to:
    :ivar correlation_id:
    :vartype correlation_id:
    :ivar content_type:
    :vartype content_type:
    :ivar content_encoding:
    :vartype content_encoding:
    :ivar absolute_expiry_time:
    :vartype absolute_expiry_time:
    :ivar creation_time:
    :vartype creation_time:
    :ivar group_id:
    :vartype group_id:
    :ivar group_sequence:
    :vartype group_sequence:
    :ivar reply_to_group_id:
    :vartype reply_to_group_id:
    """

    def __init__(self,
                 message_id=None,
                 user_id=None,
                 to=None,
                 subject=None,
                 reply_to=None,
                 correlation_id=None,
                 content_type=None,
                 content_encoding=None,
                 absolute_expiry_time=None,
                 creation_time=None,
                 group_id=None,
                 group_sequence=None,
                 reply_to_group_id=None,
                 properties=None,
                 encoding='UTF-8'):
        self._encoding = encoding
        if properties:
            self._message_id = properties.message_id
            self._user_id = properties.user_id
            self._to = properties.to
            self._subject = properties.subject
            self._reply_to = properties.reply_to
            self._correlation_id = properties.correlation_id
            self._content_type = properties.content_type
            self._content_encoding = properties.content_encoding
            self._absolute_expiry_time = properties.absolute_expiry_time
            self._creation_time = properties.creation_time
            self._group_id = properties.group_id
            self._group_sequence = properties.group_sequence
            self._reply_to_group_id = properties.reply_to_group_id
        else:
            self.message_id = message_id
            self.user_id = user_id
            self.to = to
            self.subject = subject
            self.reply_to = reply_to
            self.correlation_id = correlation_id
            self.content_type = content_type
            self.content_encoding = content_encoding
            self.absolute_expiry_time = absolute_expiry_time
            self.creation_time = creation_time
            self.group_id = group_id
            self.group_sequence = group_sequence
            self.reply_to_group_id = reply_to_group_id

    def __str__(self):
        return str({
            'message_id': self.message_id,
            'user_id': self.user_id,
            'to': self.to,
            'subject': self.subject,
            'reply_to': self.reply_to,
            'correlation_id': self.correlation_id,
            'content_type': self.content_type,
            'content_encoding': self.content_encoding,
            'absolute_expiry_time': self.absolute_expiry_time,
            'creation_time': self.creation_time,
            'group_id': self.group_id,
            'group_sequence': self.group_sequence,
            'reply_to_group_id': self.reply_to_group_id
        })

    @property
    def message_id(self):
        if self._message_id:
            return self._message_id.value
        return None

    @message_id.setter
    def message_id(self, value):
        if value is None:
            self._message_id = None
        else:
            self._message_id = utils.data_factory(value, encoding=self._encoding)

    @property
    def user_id(self):
        return self._user_id

    @user_id.setter
    def user_id(self, value):
        if isinstance(value, six.text_type):
            value = value.encode(self._encoding)
        elif value is not None and not isinstance(value, six.binary_type):
            raise TypeError("user_id must be bytes or str.")
        self._user_id = value

    @property
    def to(self):
        if self._to:
            return self._to.value
        return None

    @to.setter
    def to(self, value):
        if value is None:
            self._to = None
        else:
            self._to = utils.data_factory(value, encoding=self._encoding)

    @property
    def subject(self):
        return self._subject

    @subject.setter
    def subject(self, value):
        if isinstance(value, six.text_type):
            value = value.encode(self._encoding)
        elif value is not None and not isinstance(value, six.binary_type):
            raise TypeError("subject must be bytes or str.")
        self._subject = value

    @property
    def reply_to(self):
        if self._reply_to is not None:
            return self._reply_to.value
        return None

    @reply_to.setter
    def reply_to(self, value):
        if value is None:
            self._reply_to = None
        else:
            self._reply_to = utils.data_factory(value, encoding=self._encoding)

    @property
    def correlation_id(self):
        if self._correlation_id is not None:
            return self._correlation_id.value
        return None

    @correlation_id.setter
    def correlation_id(self, value):
        if value is None:
            self._correlation_id = None
        else:
            self._correlation_id = utils.data_factory(value, encoding=self._encoding)

    @property
    def content_type(self):
        return self._content_type

    @content_type.setter
    def content_type(self, value):
        if isinstance(value, six.text_type):
            value = value.encode(self._encoding)
        elif value is not None and not isinstance(value, six.binary_type):
            raise TypeError("content_type must be bytes or str.")
        self._content_type = value

    @property
    def content_encoding(self):
        return self._content_encoding

    @content_encoding.setter
    def content_encoding(self, value):
        if isinstance(value, six.text_type):
            value = value.encode(self._encoding)
        elif value is not None and not isinstance(value, six.binary_type):
            raise TypeError("content_encoding must be bytes or str.")
        self._content_encoding = value

    @property
    def absolute_expiry_time(self):
        return self._absolute_expiry_time

    @absolute_expiry_time.setter
    def absolute_expiry_time(self, value):
        if value is not None and not isinstance(value, int):
            raise TypeError("absolute_expiry_time must be an integer.")
        self._absolute_expiry_time = value

    @property
    def creation_time(self):
        return self._creation_time

    @creation_time.setter
    def creation_time(self, value):
        if value is not None and not isinstance(value, int):
            raise TypeError("creation_time must be an integer.")
        self._creation_time = value

    @property
    def group_id(self):
        return self._group_id

    @group_id.setter
    def group_id(self, value):
        if isinstance(value, six.text_type):
            value = value.encode(self._encoding)
        elif value is not None and not isinstance(value, six.binary_type):
            raise TypeError("group_id must be bytes or str.")
        self._group_id = value

    @property
    def group_sequence(self):
        return self._group_sequence

    @group_sequence.setter
    def group_sequence(self, value):
        if value is not None and not isinstance(value, int):
            raise TypeError("group_sequence must be an integer.")
        self._group_sequence = value

    @property
    def reply_to_group_id(self):
        return self._reply_to_group_id

    @reply_to_group_id.setter
    def reply_to_group_id(self, value):
        if isinstance(value, six.text_type):
            value = value.encode(self._encoding)
        elif value is not None and not isinstance(value, six.binary_type):
            raise TypeError("reply_to_group_id must be bytes or str.")
        self._reply_to_group_id = value

    def _set_attr(self, attr, properties):
        attr_value = getattr(self, "_" + attr)
        if attr_value is not None:
            setattr(properties, attr, attr_value)

    def get_properties_obj(self):
        """Get the underlying C reference from this object.

        :rtype: uamqp.c_uamqp.cProperties
        """
        properties = c_uamqp.cProperties()
        self._set_attr('message_id', properties)
        self._set_attr('user_id', properties)
        self._set_attr('to', properties)
        self._set_attr('subject', properties)
        self._set_attr('reply_to', properties)
        self._set_attr('correlation_id', properties)
        self._set_attr('content_type', properties)
        self._set_attr('content_encoding', properties)
        self._set_attr('absolute_expiry_time', properties)
        self._set_attr('creation_time', properties)
        self._set_attr('group_id', properties)
        self._set_attr('group_sequence', properties)
        self._set_attr('reply_to_group_id', properties)
        return properties


class MessageBody(object):
    """Base class for an AMQP message body. This should
    not be used directly.
    """

    def __init__(self, c_message, encoding='UTF-8'):
        self._message = c_message
        self._encoding = encoding

    @property
    def type(self):
        return self._message.body_type

    @property
    def data(self):
        raise NotImplementedError("Only MessageBody subclasses have data.")


class DataBody(MessageBody):
    """An AMQP message body of type Data. This represents
    a list of bytes sections.

    :ivar type: The body type. This should always be `DataType`.
    :vartype type: uamqp.c_uamqp.MessageBodyType
    :ivar data: The data contained in the message body. This returns
     a generator to iterate over each section in the body, where
     each section will be a byte string.
    :vartype data: Generator[bytes]
    """

    def __str__(self):
        if six.PY3:
            return "".join(d.decode(self._encoding) for d in self.data)
        return "".join(self.data)

    def __unicode__(self):
        return u"".join(d.decode(self._encoding) for d in self.data)

    def __bytes__(self):
        return b"".join(self.data)

    def __len__(self):
        return self._message.count_body_data()

    def __getitem__(self, index):
        if index >= len(self):
            raise IndexError("Index is out of range.")
        data = self._message.get_body_data(index)
        return data.value

    def append(self, data):
        """Append a section to the body.

        :param data: The data to append.
        :type data: str or bytes
        """
        if isinstance(data, six.text_type):
            self._message.add_body_data(data.encode(self._encoding))
        elif isinstance(data, six.binary_type):
            self._message.add_body_data(data)

    @property
    def data(self):
        for i in range(len(self)):
            yield self._message.get_body_data(i)


class ValueBody(MessageBody):
    """An AMQP message body of type Value. This represents
    a single encoded object.

    :ivar type: The body type. This should always be ValueType
    :vartype type: uamqp.c_uamqp.MessageBodyType
    :ivar data: The data contained in the message body. The value
     of the encoded object
    :vartype data: object
    """

    def __str__(self):
        data = self.data
        if not data:
            return ""
        if six.PY3 and isinstance(data, six.binary_type):
            return data.decode(self._encoding)
        return str(data)

    def __unicode__(self):
        data = self.data
        if not data:
            return u""
        if isinstance(data, six.binary_type):
            return data.decode(self._encoding)
        return unicode(data)  # pylint: disable=undefined-variable

    def __bytes__(self):
        data = self.data
        if not data:
            return b""
        return bytes(data)

    def set(self, value):
        """Set a value as the message body. This can be any
        Python data type and it will be automatically encoded
        into an AMQP type. If a specific AMQP type is required, a
        `types.AMQPType` can be used.

        :param data: The data to send in the body.
        :type data: ~uamqp.types.AMQPType
        """
        value = utils.data_factory(value)
        self._message.set_body_value(value)

    @property
    def data(self):
        _value = self._message.get_body_value()
        if _value:
            return _value.value
        return None


class MessageHeader(object):
    """The Message header. This is only used on received message, and not
    set on messages being sent. The properties set on any given message
    will depend on the Service and not all messages will have all properties.

    :ivar delivery_count: The number of unsuccessful previous attempts to deliver
     this message. If this value is non-zero it can be taken as an indication that the
     delivery might be a duplicate. On first delivery, the value is zero. It is
     incremented upon an outcome being settled at the sender, according to rules
     defined for each outcome.
    :vartype delivery_count: int
    :ivar time_to_live: Duration in milliseconds for which the message is to be considered "live".
     If this is set then a message expiration time will be computed based on the time of arrival
     at an intermediary. Messages that live longer than their expiration time will be discarded
     (or dead lettered). When a message is transmitted by an intermediary that was received
     with a ttl, the transmitted message's header SHOULD contain a ttl that is computed as the
     difference between the current time and the formerly computed message expiration time,
     i.e., the reduced ttl, so that messages will eventually die if they end up in a delivery loop.
    :vartype time_to_live: int
    :ivar durable: Durable messages MUST NOT be lost even if an intermediary is unexpectedly terminated
     and restarted. A target which is not capable of fulfilling this guarantee MUST NOT accept messages
     where the durable header is set to `True`: if the source allows the rejected outcome then the
     message SHOULD be rejected with the precondition-failed error, otherwise the link MUST be detached
     by the receiver with the same error.
    :vartype durable: bool
    :ivar first_acquirer: If this value is `True`, then this message has not been acquired
     by any other link. If this value is `False`, then this message MAY have previously
     been acquired by another link or links.
    :vartype first_acquirer: bool
    :ivar priority: This field contains the relative message priority. Higher numbers indicate higher
     priority messages. Messages with higher priorities MAY be delivered before those with lower priorities.
    :vartype priority: int

    :param header: Internal only. This is used to wrap an existing message header
     that has been received from an AMQP service.
    :type header: uamqp.c_uamqp.cHeader
    """

    def __init__(self, header=None):
        self.delivery_count = None
        self.time_to_live = None
        self.first_acquirer = None
        self.durable = None
        self.priority = None
        if header:
            self.delivery_count = header.delivery_count
            self.time_to_live = header.time_to_live
            self.first_acquirer = header.first_acquirer
            self.durable = header.durable
            self.priority = header.priority

    def __str__(self):
        return str({
            'delivery_count': self.delivery_count,
            'time_to_live': self.time_to_live,
            'first_acquirer': self.first_acquirer,
            'durable': self.durable,
            'priority': self.priority
        })

    def get_header_obj(self):
        """Get the underlying C reference from this object.

        :rtype: uamqp.c_uamqp.cHeader
        """
        header = c_uamqp.create_header()
        if self.delivery_count is not None:
            header.delivery_count = self.delivery_count
        if self.time_to_live is not None:
            header.time_to_live = self.time_to_live
        if self.first_acquirer is not None:
            header.first_acquirer = self.first_acquirer
        if self.durable is not None:
            header.durable = self.durable
        if self.priority is not None:
            header.priority = self.priority
        return header



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

from libc cimport stdint

cimport c_amqpvalue
cimport c_connection
cimport c_amqp_definitions


cdef extern from "azure_uamqp_c/session.h":

    ctypedef struct SESSION_HANDLE:
        pass
    ctypedef struct LINK_ENDPOINT_HANDLE:
        pass

    ctypedef bint (*ON_LINK_ATTACHED)(void* context, LINK_ENDPOINT_HANDLE new_link_endpoint, const char* name, c_amqp_definitions.role role, c_amqpvalue.AMQP_VALUE source, c_amqpvalue.AMQP_VALUE target, c_amqp_definitions.fields properties)

    SESSION_HANDLE session_create(c_connection.CONNECTION_HANDLE connection, ON_LINK_ATTACHED on_link_attached, void* callback_context)
    int session_set_incoming_window(SESSION_HANDLE session, stdint.uint32_t incoming_window)
    int session_get_incoming_window(SESSION_HANDLE session, stdint.uint32_t* incoming_window)
    int session_set_outgoing_window(SESSION_HANDLE session, stdint.uint32_t outgoing_window)
    int session_get_outgoing_window(SESSION_HANDLE session, stdint.uint32_t* outgoing_window)
    int session_set_handle_max(SESSION_HANDLE session, c_amqp_definitions.handle handle_max)
    int session_get_handle_max(SESSION_HANDLE session, c_amqp_definitions.handle* handle_max)
    void session_destroy(SESSION_HANDLE session)
    int session_begin(SESSION_HANDLE session)
    int session_end(SESSION_HANDLE session, const char* condition_value, const char* description)



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

import logging

from uamqp import constants, errors, session
from uamqp.utils import get_running_loop
from uamqp.async_ops.mgmt_operation_async import MgmtOperationAsync

_logger = logging.getLogger(__name__)


class SessionAsync(session.Session):
    """An asynchronous AMQP Session. A Connection can have multiple Sessions, and each
    Session can have multiple Links.
    An async Session must be used with an async Connection.

    :ivar incoming_window: The size of the allowed window for incoming messages.
    :vartype incoming_window: int
    :ivar outgoing_window: The size of the allowed window for outgoing messages.
    :vartype outgoing_window: int
    :ivar handle_max: The maximum number of concurrent link handles.
    :vartype handle_max: int

    :param connection: The underlying Connection for the Session.
    :type connection: ~uamqp.async_ops.connection_async.ConnectionAsync
    :param incoming_window: The size of the allowed window for incoming messages.
    :type incoming_window: int
    :param outgoing_window: The size of the allowed window for outgoing messages.
    :type outgoing_window: int
    :param handle_max: The maximum number of concurrent link handles.
    :type handle_max: int
    :param on_attach: A callback function to be run on receipt of an ATTACH frame.
     The function must take 4 arguments: source, target, properties and error.
    :type on_attach: func[~uamqp.address.Source, ~uamqp.address.Target, dict, ~uamqp.errors.AMQPConnectionError]
    :param loop: A user specified event loop.
    :type loop: ~asycnio.AbstractEventLoop
    """

    def __init__(self, connection,
                 incoming_window=None,
                 outgoing_window=None,
                 handle_max=None,
                 on_attach=None,
                 loop=None):
        self.loop = loop or get_running_loop()
        super(SessionAsync, self).__init__(
            connection,
            incoming_window=incoming_window,
            outgoing_window=outgoing_window,
            handle_max=handle_max,
            on_attach=on_attach)

    async def __aenter__(self):
        """Run Session in an async context manager."""
        return self

    async def __aexit__(self, *args):
        """Close and destroy sesion on exiting  an async context manager."""
        await self.destroy_async()

    async def mgmt_request_async(self, message, operation, op_type=None, node=b'$management', **kwargs):
        """Asynchronously run a request/response operation. These are frequently used
        for management tasks against a $management node, however any node name can be
        specified and the available options will depend on the target service.

        :param message: The message to send in the management request.
        :type message: ~uamqp.message.Message
        :param operation: The type of operation to be performed. This value will
         be service-specific, but common values include READ, CREATE and UPDATE.
         This value will be added as an application property on the message.
        :type operation: bytes or str
        :param op_type: The type on which to carry out the operation. This will
         be specific to the entities of the service. This value will be added as
         an application property on the message.
        :type op_type: bytes or str
        :param node: The target node. Default is `b"$management"`.
        :type node: bytes or str
        :param timeout: Provide an optional timeout in milliseconds within which a response
         to the management request must be received.
        :type timeout: float
        :param status_code_field: Provide an alternate name for the status code in the
         response body which can vary between services due to the spec still being in draft.
         The default is `b"statusCode"`.
        :type status_code_field: bytes or str
        :param description_fields: Provide an alternate name for the description in the
         response body which can vary between services due to the spec still being in draft.
         The default is `b"statusDescription"`.
        :type description_fields: bytes or str
        :param encoding: The encoding to use for parameters supplied as strings.
         Default is 'UTF-8'
        :type encoding: str
        :rtype: ~uamqp.message.Message
        """
        timeout = kwargs.pop('timeout', None) or 0
        parse_response = kwargs.pop('callback', None)
        try:
            mgmt_link = self._mgmt_links[node]
        except KeyError:
            mgmt_link = MgmtOperationAsync(self, target=node, loop=self.loop, **kwargs)
            self._mgmt_links[node] = mgmt_link
            while not mgmt_link.open and not mgmt_link.mgmt_error:
                await self._connection.work_async()
            if mgmt_link.mgmt_error:
                raise mgmt_link.mgmt_error
            if mgmt_link.open != constants.MgmtOpenStatus.Ok:
                raise errors.AMQPConnectionError("Failed to open mgmt link: {}".format(mgmt_link.open))
        op_type = op_type or b'empty'
        status, response, description = await mgmt_link.execute_async(operation, op_type, message, timeout=timeout)
        if parse_response:
            return parse_response(status, response, description)
        return response

    async def destroy_async(self):
        """Asynchronously close any open management Links and close the Session.
        Cleans up and C objects for both mgmt Links and Session.
        """
        for _, link in self._mgmt_links.items():
            await link.destroy_async()
        self._session.destroy()



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

import logging

from uamqp import constants, errors, receiver
from uamqp.utils import get_running_loop

_logger = logging.getLogger(__name__)


class MessageReceiverAsync(receiver.MessageReceiver):
    """An asynchronous Message Receiver that opens its own exclsuive Link on an
    existing Session.

    :ivar receive_settle_mode: The mode by which to settle message receive
     operations. If set to `PeekLock`, the receiver will lock a message once received until
     the client accepts or rejects the message. If set to `ReceiveAndDelete`, the service
     will assume successful receipt of the message and clear it from the queue. The
     default is `PeekLock`.
    :vartype receive_settle_mode: ~uamqp.constants.ReceiverSettleMode
    :ivar send_settle_mode: The mode by which to settle message send
     operations. If set to `Unsettled`, the client will wait for a confirmation
     from the service that the message was successfully sent. If set to 'Settled',
     the client will not wait for confirmation and assume success.
    :vartype send_settle_mode: ~uamqp.constants.SenderSettleMode
    :ivar max_message_size: The maximum allowed message size negotiated for the Link.
    :vartype max_message_size: int

    :param session: The underlying Session with which to receive.
    :type session: ~uamqp.session.Session
    :param source: The AMQP endpoint to receive from.
    :type source: ~uamqp.address.Source
    :param target: The name of target (i.e. the client).
    :type target: str or bytes
    :param name: A unique name for the receiver. If not specified a GUID will be used.
    :type name: str or bytes
    :param receive_settle_mode: The mode by which to settle message receive
     operations. If set to `PeekLock`, the receiver will lock a message once received until
     the client accepts or rejects the message. If set to `ReceiveAndDelete`, the service
     will assume successful receipt of the message and clear it from the queue. The
     default is `PeekLock`.
    :type receive_settle_mode: ~uamqp.constants.ReceiverSettleMode
    :param send_settle_mode: The mode by which to settle message send
     operations. If set to `Unsettled`, the client will wait for a confirmation
     from the service that the message was successfully sent. If set to 'Settled',
     the client will not wait for confirmation and assume success.
    :type send_settle_mode: ~uamqp.constants.SenderSettleMode
    :param max_message_size: The maximum allowed message size negotiated for the Link.
    :type max_message_size: int
    :param prefetch: The receiver Link credit that determines how many
     messages the Link will attempt to handle per connection iteration.
    :type prefetch: int
    :param properties: Metadata to be sent in the Link ATTACH frame.
    :type properties: dict
    :param error_policy: A policy for parsing errors on link, connection and message
     disposition to determine whether the error should be retryable.
    :type error_policy: ~uamqp.errors.ErrorPolicy
    :param debug: Whether to turn on network trace logs. If `True`, trace logs
     will be logged at INFO level. Default is `False`.
    :type debug: bool
    :param encoding: The encoding to use for parameters supplied as strings.
     Default is 'UTF-8'
    :type encoding: str
    :param loop: A user specified event loop.
    :type loop: ~asycnio.AbstractEventLoop
    """

    def __init__(self, session, source, target,
                 on_message_received,
                 name=None,
                 receive_settle_mode=constants.ReceiverSettleMode.PeekLock,
                 send_settle_mode=constants.SenderSettleMode.Unsettled,
                 max_message_size=constants.MAX_MESSAGE_LENGTH_BYTES,
                 prefetch=300,
                 properties=None,
                 error_policy=None,
                 debug=False,
                 encoding='UTF-8',
                 loop=None):
        self.loop = loop or get_running_loop()
        super(MessageReceiverAsync, self).__init__(
            session, source, target,
            on_message_received,
            name=name,
            receive_settle_mode=receive_settle_mode,
            send_settle_mode=send_settle_mode,
            max_message_size=max_message_size,
            prefetch=prefetch,
            properties=properties,
            error_policy=error_policy,
            debug=debug,
            encoding=encoding)

    async def __aenter__(self):
        """Open the MessageReceiver in an async context manager."""
        await self.open_async()
        return self

    async def __aexit__(self, *args):
        """Close the MessageReceiver when exiting an async context manager."""
        await self.destroy_async()

    async def destroy_async(self):
        """Asynchronously close both the Receiver and the Link. Clean up any C objects."""
        self.destroy()

    async def open_async(self):
        """Asynchronously open the MessageReceiver in order to start
        processing messages.

        :raises: ~uamqp.errors.AMQPConnectionError if the Receiver raises
         an error on opening. This can happen if the source URI is invalid
         or the credentials are rejected.
        """
        try:
            self._receiver.open(self)
        except ValueError:
            raise errors.AMQPConnectionError(
                "Failed to open Message Receiver. "
                "Please confirm credentials and target URI.")

    async def close_async(self):
        """Close the Receiver asynchronously, leaving the link intact."""
        self.close()



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

import os
import sys
import re
import distutils
import subprocess
import platform
from setuptools import find_packages, monkey, setup, Extension
from setuptools.command.build_ext import build_ext as build_ext_orig
from distutils.extension import Extension
from distutils.version import LooseVersion
from distutils import log as logger

try:
    from Cython.Build import cythonize
    USE_CYTHON = True
except ImportError:
    USE_CYTHON = False

# If the C file doesn't exist and no Cython is available, die
if not os.path.exists("uamqp/c_uamqp.c") and not USE_CYTHON:
    raise ValueError("You need to install cython==0.27.3 in order to execute this setup.py if 'uamqp/c_uamqp.c' does not exists")

is_27 = sys.version_info < (3,)
is_x64 = platform.architecture()[0] == '64bit'
is_win = sys.platform.startswith('win')
is_mac = sys.platform.startswith('darwin')
rebuild_pyx = os.environ.get('UAMQP_REBUILD_PYX', False)
use_openssl = not (is_win or is_mac)
#  use_openssl = os.environ.get('UAMQP_USE_OPENSSL', not (is_win or is_mac))
supress_link_flags = os.environ.get("UAMQP_SUPPRESS_LINK_FLAGS", False)

# Version extraction inspired from 'requests'
with open(os.path.join('uamqp', '__init__.py'), 'r') as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        fd.read(), re.MULTILINE).group(1)


cwd = os.path.abspath('.')

# Headers

pxd_inc_dir = os.path.join(cwd, "src", "vendor", "inc")
sys.path.insert(0, pxd_inc_dir)

include_dirs = [
    pxd_inc_dir,
    # azure-c-shared-utility inc
    "./src/vendor/azure-uamqp-c/deps/azure-c-shared-utility/pal/inc",
    "./src/vendor/azure-uamqp-c/deps/azure-c-shared-utility/inc",
    # azure-uamqp-c inc
    "./src/vendor/azure-uamqp-c/inc",
]
if is_win:
    include_dirs.append("./src/vendor/azure-uamqp-c/deps/azure-c-shared-utility/pal/windows")
    if is_27:
        include_dirs.append("./src/vendor/azure-uamqp-c/deps/azure-c-shared-utility/inc/azure_c_shared_utility/windowsce")
else:
    include_dirs.append("./src/vendor/azure-uamqp-c/deps/azure-c-shared-utility/pal/linux")


# Build unique source pyx

def create_cython_file():
    content_includes = ""
    for f in os.listdir("./src"):
        if f.endswith(".pyx"):
            print("Adding {}".format(f))
            content_includes += "include \"src/" + f + "\"\n"
    c_uamqp_src = os.path.join("uamqp", "c_uamqp.pyx")
    with open(c_uamqp_src, 'w') as lib_file:
        lib_file.write(content_includes)
    return c_uamqp_src


def get_build_env():
    build_env = os.environ.copy()
    return {k.upper(): v for k, v in build_env.items()}

def is_msvc_9_for_python_compiler():
    return is_win and is_27 and os.path.exists("C:\\Program Files (x86)\\Common Files\\Microsoft\\Visual C++ for Python\\9.0")

def get_generator():
    if is_msvc_9_for_python_compiler():
        return "NMake Makefiles"
    if is_win:
        generator = "Visual Studio 9 2008" if is_27 else "Visual Studio 15 2017"
        if is_x64:
            return generator + " Win64"
        else:
            return generator
    return "Unix Makefiles"


def get_latest_windows_sdk():
    """The Windows SDK that ships with VC++ 9.0 is not new enough to include schannel support.
    So we need to check the OS to see if a newer Windows SDK is available to link to.

    This is only run on Windows if using Python 2.7.
    """

    from _winreg import ConnectRegistry, OpenKey, EnumKey, QueryValueEx, HKEY_LOCAL_MACHINE
    installed_sdks = {}
    key_path = "SOFTWARE\\Wow6432Node\\Microsoft\\Microsoft SDKs\\Windows" if is_x64 else \
        "SOFTWARE\\Microsoft\\Microsoft SDKs\\Windows"
    try:
        with ConnectRegistry(None, HKEY_LOCAL_MACHINE) as hklm:
            with OpenKey(hklm, key_path) as handler:
                for i in range(1024):
                    try:
                        asubkey_name=EnumKey(handler, i)
                        with OpenKey(handler, asubkey_name) as asubkey:
                            location = QueryValueEx(asubkey, "InstallationFolder")
                            if not location or not os.path.isdir(location[0]):
                                continue
                            version = QueryValueEx(asubkey, "ProductVersion")
                            if not version:
                                continue
                            installed_sdks[version[0].lstrip('v')] = location[0]
                            logger.info("Found Windows SDK %s at %s" % (version[0], location[0]))
                    except EnvironmentError:
                        break
    except EnvironmentError:
        print("Warning - build may fail: No Windows SDK found.")
        return []

    installed_versions = sorted([LooseVersion(k) for k in installed_sdks.keys()])
    if installed_versions[-1] < LooseVersion("8.1"):
        print("Warning - build may fail: Cannot find Windows SDK 8.1 or greater.")
        return []

    latest_sdk_version = installed_versions[-1].vstring
    logger.info("Selecting Windows SDK v%s" % latest_sdk_version)
    lib_path = os.path.join(installed_sdks[latest_sdk_version], "lib")
    libs_dirs = [d for d in os.listdir(lib_path) if os.path.isdir(os.path.join(lib_path, d, "um"))]
    if not libs_dirs:
        print("Warning - build may fail: No Windows SDK libraries found.")
        return []

    logger.info("Found SDK libraries %s in Windows SDK v%s" % (", ".join(libs_dirs), latest_sdk_version))
    latest_libs_dir = max([LooseVersion(d) for d in libs_dirs]).vstring
    logger.info("Selecting SDK libraries %s" % latest_libs_dir)

    platform_libs_path = os.path.join(lib_path, latest_libs_dir, "um", 'x64' if is_x64 else 'x86')
    if not os.path.isdir(platform_libs_path):
        print("Warning - build may fail: Windows SDK v{} libraries {} not found at path {}.".format(
            latest_sdk_version, latest_libs_dir, platform_libs_path))
    else:
        print("Adding Windows SDK v{} libraries {} to search path, installed at {}".format(
            latest_sdk_version, latest_libs_dir, platform_libs_path))

    return [platform_libs_path]

def get_msvc_env(vc_ver):
    arch = "amd64" if is_x64 else "x86"
    msvc_env = distutils.msvc9compiler.query_vcvarsall(vc_ver, arch)
    return {str(k.upper()): str(v) for k, v in msvc_env.items()}


# Compile uamqp
# Inspired by https://stackoverflow.com/a/48015772/4074838

class UAMQPExtension(Extension):

    def __init__(self, name):
        # don't invoke the original build_ext for this special extension
        Extension.__init__(self, name, sources=[])

def create_folder_no_exception(foldername):
    try:
        os.makedirs(foldername)
    except Exception: # Assume it's already there, and not impossible to create
        pass

class build_ext(build_ext_orig):

    def run(self):
        monkey.patch_all()
        cmake_build_dir = None
        for ext in self.extensions:
            if isinstance(ext, UAMQPExtension):
                self.build_cmake(ext)
                # Now I have built in ext.cmake_build_dir
                cmake_build_dir = self.cmake_build_dir
            else:
                ext.library_dirs += [
                    cmake_build_dir,
                    cmake_build_dir + "/deps/azure-c-shared-utility/",
                    cmake_build_dir + "/Debug/",
                    cmake_build_dir + "/Release/",
                    cmake_build_dir + "/deps/azure-c-shared-utility/Debug/",
                    cmake_build_dir + "/deps/azure-c-shared-utility/Release/"
                ]

        if is_win and is_27:
            ext.library_dirs.extend([lib for lib in get_msvc_env(9.0)['LIB'].split(';') if lib])
            ext.library_dirs.extend(get_latest_windows_sdk())

        build_ext_orig.run(self)

    def build_cmake(self, ext):
        cwd = os.getcwd()

        # these dirs will be created in build_py, so if you don't have
        # any python sources to bundle, the dirs will be missing
        self.cmake_build_dir = self.build_temp + "/cmake"
        create_folder_no_exception(self.cmake_build_dir)

        extdir = self.get_ext_fullpath(ext.name)
        create_folder_no_exception(extdir)

        logger.info("will build uamqp in %s", self.cmake_build_dir)
        os.chdir(cwd + "/" + self.cmake_build_dir)

        generator = get_generator()
        logger.info("Building with generator: {}".format(generator))

        build_env = get_build_env()
        if is_msvc_9_for_python_compiler():
            logger.info("Configuring environment for Microsoft Visual C++ Compiler for Python 2.7")
            build_env.update(get_msvc_env(9.0))

        # Configure
        configure_command = [
            "cmake",
            cwd + "/src/vendor/azure-uamqp-c/",
            "-G \"{}\"".format(generator),
            "-Duse_openssl:bool={}".format("ON" if use_openssl else "OFF"),
            "-Duse_default_uuid:bool=ON ", # Should we use libuuid in the system or light one?
            "-Duse_builtin_httpapi:bool=ON ", # Should we use libcurl in the system or light one?
            "-Dskip_samples:bool=ON", # Don't compile uAMQP samples binaries
            "-DCMAKE_POSITION_INDEPENDENT_CODE=TRUE", # ask for -fPIC
            "-DCMAKE_BUILD_TYPE=Release"
        ]

        joined_cmd = " ".join(configure_command)
        logger.info("calling %s", joined_cmd)
        subprocess.check_call(joined_cmd, shell=True, universal_newlines=True, env=build_env)

        compile_command  = ["cmake", "--build", ".", "--config", "Release"]
        joined_cmd = " ".join(compile_command)
        logger.info("calling %s", joined_cmd)
        subprocess.check_call(joined_cmd, shell=True, universal_newlines=True, env=build_env)

        os.chdir(cwd)

        if USE_CYTHON:
            create_cython_file()

# Libraries and extra compile args

kwargs = {}
if is_win:
    kwargs['libraries'] = [
        'uamqp',
        'aziotsharedutil',
        'AdvAPI32',
        'Crypt32',
        'ncrypt',
        'Secur32',
        'schannel',
        'RpcRT4',
        'WSock32',
        'WS2_32']
elif is_mac:
    kwargs['extra_compile_args'] = ['-g', '-O0', "-std=gnu99", "-fPIC"]
    kwargs['libraries'] = ['uamqp', 'aziotsharedutil']
    if use_openssl and not supress_link_flags:
        kwargs['libraries'].extend(['azssl', 'azcrypto'])
    elif not use_openssl:
        kwargs['extra_link_args'] = [
            '-framework', 'CoreFoundation',
            '-framework', 'CFNetwork',
            '-framework', 'Security']
else:
    kwargs['extra_compile_args'] = ['-g', '-O0', "-std=gnu99", "-fPIC"]
    # SSL before crypto matters: https://bugreports.qt.io/browse/QTBUG-62692
    kwargs['libraries'] = ['uamqp', 'aziotsharedutil']
    if sys.version_info < (3, 5):
        kwargs['libraries'].append('rt')
    if not supress_link_flags:
        kwargs['libraries'].extend(['ssl', 'crypto'])


# If the C file doesn't exist, build the "c_uamqp.c" file
# That's not perfect, since we build it on a "--help", but should be if cloned from source only.
c_uamqp_src = "uamqp/c_uamqp.c"
if not os.path.exists(c_uamqp_src) or rebuild_pyx:
    c_uamqp_src = create_cython_file()

sources = [
    c_uamqp_src
]

# Configure the extension

extensions = [
    UAMQPExtension("cmake_uamqp"),
    Extension(
        "uamqp.c_uamqp",
        sources=sources,
        include_dirs=include_dirs,
        **kwargs
    )
]

with open('README.rst') as f:  # , encoding='utf-8'
    readme = f.read()
with open('HISTORY.rst') as f:  # , encoding='utf-8'
    history = f.read()

if USE_CYTHON:
    extensions = cythonize(extensions)

setup(
    name='uamqp',
    version=version,
    description='AMQP 1.0 Client Library for Python',
    long_description=readme + '\n\n' + history,
    license='MIT License',
    author='Microsoft Corporation',
    author_email='azpysdkhelp@microsoft.com',
    url='https://github.com/Azure/azure-uamqp-python',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Cython',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'License :: OSI Approved :: MIT License'
    ],
    zip_safe=False,
    include_package_data=True,
    packages=find_packages(exclude=["tests"]),
    ext_modules=extensions,
    install_requires=[
        "certifi>=2017.4.17",
        "six~=1.0"
    ],
    extras_require={
        ":python_version<'3.4'": ['enum34>=1.0.4']
    },
    cmdclass={
        'build_ext': build_ext,
    },
    python_requires=">=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*",
)



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

from .connection_async import ConnectionAsync
from .session_async import SessionAsync
from .sender_async import MessageSenderAsync
from .receiver_async import MessageReceiverAsync
from .mgmt_operation_async import MgmtOperationAsync



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

from libc cimport stdint

cimport c_amqp_definitions
cimport c_amqpvalue
cimport c_session
cimport c_async_operation

cdef extern from "azure_uamqp_c/link.h":

    ctypedef struct LINK_HANDLE:
        pass

    ctypedef struct ON_LINK_DETACH_EVENT_SUBSCRIPTION_HANDLE:
        pass

    cdef enum LINK_STATE_TAG:
        LINK_STATE_DETACHED,
        LINK_STATE_HALF_ATTACHED_ATTACH_SENT,
        LINK_STATE_HALF_ATTACHED_ATTACH_RECEIVED,
        LINK_STATE_ATTACHED,
        LINK_STATE_ERROR

    cdef enum LINK_TRANSFER_RESULT_TAG:
        LINK_TRANSFER_ERROR,
        LINK_TRANSFER_BUSY

    cdef enum LINK_DELIVERY_SETTLE_REASON_TAG:
        LINK_DELIVERY_SETTLE_REASON_DISPOSITION_RECEIVED,
        LINK_DELIVERY_SETTLE_REASON_SETTLED,
        LINK_DELIVERY_SETTLE_REASON_NOT_DELIVERED,
        LINK_DELIVERY_SETTLE_REASON_TIMEOUT,
        LINK_DELIVERY_SETTLE_REASON_CANCELLED

    ctypedef void (*ON_LINK_DETACH_RECEIVED)(void* context, c_amqp_definitions.ERROR_HANDLE error)

    LINK_HANDLE link_create(c_session.SESSION_HANDLE session, const char* name, c_amqp_definitions.role role, c_amqpvalue.AMQP_VALUE source, c_amqpvalue.AMQP_VALUE target)
    void link_destroy(LINK_HANDLE handle)
    int link_set_snd_settle_mode(LINK_HANDLE link, c_amqp_definitions.sender_settle_mode snd_settle_mode)
    int link_get_snd_settle_mode(LINK_HANDLE link, c_amqp_definitions.sender_settle_mode* snd_settle_mode)
    int link_set_rcv_settle_mode(LINK_HANDLE link, c_amqp_definitions.receiver_settle_mode rcv_settle_mode)
    int link_get_rcv_settle_mode(LINK_HANDLE link, c_amqp_definitions.receiver_settle_mode* rcv_settle_mode)
    int link_set_initial_delivery_count(LINK_HANDLE link, c_amqp_definitions.sequence_no initial_delivery_count)
    int link_get_initial_delivery_count(LINK_HANDLE link, c_amqp_definitions.sequence_no* initial_delivery_count)
    int link_set_max_link_credit(LINK_HANDLE link, stdint.uint32_t max_link_credit)
    int link_set_max_message_size(LINK_HANDLE link, stdint.uint64_t max_message_size)
    int link_get_max_message_size(LINK_HANDLE link, stdint.uint64_t* max_message_size)
    int link_get_peer_max_message_size(LINK_HANDLE link, stdint.uint64_t* peer_max_message_size)
    int link_set_attach_properties(LINK_HANDLE link, c_amqp_definitions.fields attach_properties)
    int link_get_name(LINK_HANDLE link, const char** link_name)
    int link_get_received_message_id(LINK_HANDLE link, c_amqp_definitions.delivery_number* message_id)

    ON_LINK_DETACH_EVENT_SUBSCRIPTION_HANDLE link_subscribe_on_link_detach_received(LINK_HANDLE link, ON_LINK_DETACH_RECEIVED on_link_detach_received, void* context)
    void link_unsubscribe_on_link_detach_received(ON_LINK_DETACH_EVENT_SUBSCRIPTION_HANDLE event_subscription)



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

from libc cimport stdint

cimport c_message
cimport c_link
cimport c_async_operation
cimport c_amqp_definitions
cimport c_amqpvalue


cdef extern from "azure_uamqp_c/message_sender.h":

    cdef enum MESSAGE_SEND_RESULT_TAG:
        MESSAGE_SEND_OK,
        MESSAGE_SEND_ERROR,
        MESSAGE_SEND_TIMEOUT,
        MESSAGE_SEND_CANCELLED

    cdef enum MESSAGE_SENDER_STATE_TAG:
        MESSAGE_SENDER_STATE_IDLE,
        MESSAGE_SENDER_STATE_OPENING,
        MESSAGE_SENDER_STATE_OPEN,
        MESSAGE_SENDER_STATE_CLOSING,
        MESSAGE_SENDER_STATE_ERROR

    ctypedef struct MESSAGE_SENDER_HANDLE:
        pass

    ctypedef void (*ON_MESSAGE_SEND_COMPLETE)(void* context, MESSAGE_SEND_RESULT_TAG send_result, c_amqpvalue.AMQP_VALUE delivery_state)
    ctypedef void (*ON_MESSAGE_SENDER_STATE_CHANGED)(void* context, MESSAGE_SENDER_STATE_TAG new_state, MESSAGE_SENDER_STATE_TAG previous_state)

    MESSAGE_SENDER_HANDLE messagesender_create(c_link.LINK_HANDLE link, ON_MESSAGE_SENDER_STATE_CHANGED on_message_sender_state_changed, void* context)
    void messagesender_destroy(MESSAGE_SENDER_HANDLE message_sender)
    int messagesender_open(MESSAGE_SENDER_HANDLE message_sender)
    int messagesender_close(MESSAGE_SENDER_HANDLE message_sender)
    c_async_operation.ASYNC_OPERATION_HANDLE messagesender_send_async(MESSAGE_SENDER_HANDLE message_sender, c_message.MESSAGE_HANDLE message, ON_MESSAGE_SEND_COMPLETE on_message_send_complete, void* callback_context, c_amqp_definitions.tickcounter_ms_t timeout)
    void messagesender_set_trace(MESSAGE_SENDER_HANDLE message_sender, bint traceOn)



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

import logging

from uamqp import c_uamqp, constants, errors, mgmt_operation
from uamqp.address import Source, Target

_logger = logging.getLogger(__name__)


class Session(object):
    """An AMQP Session. A Connection can have multiple Sessions, and each
    Session can have multiple Links.

    :ivar incoming_window: The size of the allowed window for incoming messages.
    :vartype incoming_window: int
    :ivar outgoing_window: The size of the allowed window for outgoing messages.
    :vartype outgoing_window: int
    :ivar handle_max: The maximum number of concurrent link handles.
    :vartype handle_max: int

    :param connection: The underlying Connection for the Session.
    :type connection: ~uamqp.connection.Connection
    :param incoming_window: The size of the allowed window for incoming messages.
    :type incoming_window: int
    :param outgoing_window: The size of the allowed window for outgoing messages.
    :type outgoing_window: int
    :param handle_max: The maximum number of concurrent link handles.
    :type handle_max: int
    :param on_attach: A callback function to be run on receipt of an ATTACH frame.
     The function must take 4 arguments: source, target, properties and error.
    :type on_attach: func[~uamqp.address.Source, ~uamqp.address.Target, dict, ~uamqp.errors.AMQPConnectionError]
    """

    def __init__(self, connection,
                 incoming_window=None,
                 outgoing_window=None,
                 handle_max=None,
                 on_attach=None):
        self._connection = connection
        self._conn = connection._conn  # pylint: disable=protected-access
        self._session = c_uamqp.create_session(self._conn, self)
        self._mgmt_links = {}
        self._link_error = None
        self._on_attach = on_attach

        if incoming_window:
            self.incoming_window = incoming_window
        if outgoing_window:
            self.outgoing_window = outgoing_window
        if handle_max:
            self.handle_max = handle_max

    def __enter__(self):
        """Run Session in a context manager."""
        return self

    def __exit__(self, *args):
        """Close and destroy sesion on exiting context manager."""
        self._session.destroy()

    def _attach_received(self, source, target, properties, error=None):
        if error:
            self._link_error = errors.AMQPConnectionError(error)
        if self._on_attach:
            if source and target:
                source = Source.from_c_obj(source)
                target = Target.from_c_obj(target)
            if properties:
                properties = properties.value
            self._on_attach(source, target, properties, self._link_error)

    def mgmt_request(self, message, operation, op_type=None, node=b'$management', **kwargs):
        """Run a request/response operation. These are frequently used for management
        tasks against a $management node, however any node name can be specified
        and the available options will depend on the target service.

        :param message: The message to send in the management request.
        :type message: ~uamqp.message.Message
        :param operation: The type of operation to be performed. This value will
         be service-specific, but common values include READ, CREATE and UPDATE.
         This value will be added as an application property on the message.
        :type operation: bytes
        :param op_type: The type on which to carry out the operation. This will
         be specific to the entities of the service. This value will be added as
         an application property on the message.
        :type op_type: bytes
        :param node: The target node. Default is `b"$management"`.
        :type node: bytes
        :param timeout: Provide an optional timeout in milliseconds within which a response
         to the management request must be received.
        :type timeout: float
        :param status_code_field: Provide an alternate name for the status code in the
         response body which can vary between services due to the spec still being in draft.
         The default is `b"statusCode"`.
        :type status_code_field: bytes
        :param description_fields: Provide an alternate name for the description in the
         response body which can vary between services due to the spec still being in draft.
         The default is `b"statusDescription"`.
        :type description_fields: bytes
        :param encoding: The encoding to use for parameters supplied as strings.
         Default is 'UTF-8'
        :type encoding: str
        :rtype: ~uamqp.message.Message
        """
        timeout = kwargs.pop('timeout', None) or 0
        parse_response = kwargs.pop('callback', None)
        try:
            mgmt_link = self._mgmt_links[node]
        except KeyError:
            mgmt_link = mgmt_operation.MgmtOperation(self, target=node, **kwargs)
            self._mgmt_links[node] = mgmt_link
            while not mgmt_link.open and not mgmt_link.mgmt_error:
                self._connection.work()
            if mgmt_link.mgmt_error:
                raise mgmt_link.mgmt_error
            if mgmt_link.open != constants.MgmtOpenStatus.Ok:
                raise errors.AMQPConnectionError("Failed to open mgmt link: {}".format(mgmt_link.open))
        op_type = op_type or b'empty'
        status, response, description = mgmt_link.execute(operation, op_type, message, timeout=timeout)
        if parse_response:
            return parse_response(status, response, description)
        return response

    def destroy(self):
        """Close any open management Links and close the Session.
        Cleans up and C objects for both mgmt Links and Session.
        """
        for _, link in self._mgmt_links.items():
            link.destroy()
        self._session.destroy()

    @property
    def incoming_window(self):
        return self._session.incoming_window

    @incoming_window.setter
    def incoming_window(self, value):
        self._session.incoming_window = int(value)

    @property
    def outgoing_window(self):
        return self._session.outgoing_window

    @outgoing_window.setter
    def outgoing_window(self, value):
        self._session.outgoing_window = int(value)

    @property
    def handle_max(self):
        return self._session.handle_max

    @handle_max.setter
    def handle_max(self, value):
        self._session.handle_max = int(value)
